from collections.abc import Iterator, Sequence
from typing import (
    Any,
    ClassVar,
    Self,
    SupportsInt,
    dataclass_transform,
    get_origin,
    get_type_hints,
)

import numpy as np

from ._constants import (
    TWINCAT_STRING_ENCODING,
    AdsDataType,
    CommandId,
    ErrorCode,
    IndexGroup,
    StateFlag,
    SymbolFlag,
    SystemServiceCommandId,
    TransmissionMode,
)
from ._types import (
    BYTES6,
    BYTES12,
    BYTES16,
    UINT8,
    UINT16,
    UINT32,
    UINT64,
    AdsMessageDataType,
    AmsNetId,
)

# ===================================================================
# ===== THE MESSAGING FRAMEWORK
# ===================================================================


def _get_field_values(
    cls, fields: Sequence[str], kwargs: dict[str, Any]
) -> Iterator[Any]:
    """
    :param cls: the Message object type to extract values from
    :param fields: the names of the various fields which characterize this Message \
        object
    :param kwargs: map of available fields and their associated values which define \
        this Message object

    :returns: a generator of the values of the fields in the Message object

    :raises KeyError: exception arising when a required field was expected \
        but not found in the Message data structure
    """
    for field in fields:
        # Try first from kwargs
        value = kwargs.get(field, None)
        if value is None:
            # If not get it from class defaults
            value = cls.__dict__.get(field, None).default
        if value is None:
            # It was required but not passed
            raise KeyError(f"{field} is a required argument")
        yield value


def _get_dtype_arg(annotation: Any) -> Any:
    """
    Get the numpy data type from the given annotation.

    :param annotation: the type annotation to extract the numpy data type from

    :returns: the numpy data type

    :raises TypeError: exception arising when the annotation isn't a valid numpy type
    """
    if get_origin(annotation) is AdsMessageDataType:
        # If we have AdsMessageDataType[np_type, coercible_type], extract np_type
        return AdsMessageDataType.get_dtype(annotation)
    elif isinstance(annotation, type) and issubclass(annotation, np.generic):
        # If we have a subclass of a numpy generic that will do
        return annotation

    raise TypeError(annotation)


@dataclass_transform(kw_only_default=True)
class Message:
    """
    Define a generic ADS message type which the various message structures conform to.

    Instance attributes:
        _value: numpy NDArray based on the specific structure of the ADS message type

    :raises TypeError: exception arising when trying to instantiate a Message object \
        with both buffer and kwargs.
    """

    data: bytes
    """Array of bytes representing the data associated with the ADS message"""
    dtype: ClassVar[np.dtype]
    """Numpy data type representing the ADS Message."""

    def __init__(self, buffer: bytes = b"", *, data: bytes = b"", **kwargs):
        if buffer:
            if kwargs:
                raise TypeError(
                    "Can't have a Message class instantiated with both buffer "
                    "and kwargs."
                )
            self._value = np.frombuffer(buffer, self.dtype, count=1)
            self.data = buffer[self._value.nbytes :]
        else:
            fields = self.dtype.fields
            values = (
                tuple(_get_field_values(type(self), list(fields.keys()), kwargs))
                if fields is not None
                else ()
            )
            self._value = np.array([values], dtype=self.dtype)
            self.data = data

    def __getattr__(self, name: str) -> Any:
        """
        Overriding method used to access the value of the Message object attributes.
        """
        return self._value[name][0]

    def __init_subclass__(cls):
        """
        Get the type of the Message object as a numpy data type.
        It includes all fields specific to that Message, except for the 'data' field.
        Its value is computed once and then cached as a normal attribute \
            for the life of the instance.

        :returns: the ADS message data type
        """
        items = []
        for name, annotation in get_type_hints(cls).items():
            if name not in {"data", "dtype"}:
                items.append((name, _get_dtype_arg(annotation)))
        cls.dtype = np.dtype(items)

    @classmethod
    def from_bytes(cls, buffer: bytes) -> Self:
        """
        Create a Message object whose value is a numpy NDArray defined from the given \
            array of bytes.

        :param buffer: the array of bytes characterising the type of Message

        :returns: an instance of the Message class
        """
        return cls(buffer)

    def to_bytes(self, include_data: bool = True) -> bytes:
        """
        Convert a Message object into an array of bytes.

        :returns: a byte array representing the ADS message and its associated data
        """
        return (
            (self._value.tobytes() + self.data)
            if include_data
            else self._value.tobytes()
        )


class MessageRequest(Message):
    """Message interface for an ADS request to the server."""

    ...


class MessageResponse(Message):
    """Message interface for an ADS response from the server."""

    ...


# ===================================================================
# ===== UDP COMMUNICATIONS
# ===================================================================


class UDPInfo(Message):
    """Define the UDP info data structure used in ADS UDP communications."""

    tag_id: UINT16
    """The tag identifier"""
    length: UINT16
    """The length of the data in bytes"""
    data: bytes
    """The UDP info data"""


class AdsUDPMessage(Message):
    """Define the ADS UDP message structure."""

    udp_cookie: UINT32
    """Magic cookie identifying an ADS UDP message"""
    invoke_id: UINT32
    """Id used to map a received response to a udp command"""
    service_id: UINT32
    """Id used to map a command/response to a udp service"""
    data: bytes
    """The UDP message data"""

    @classmethod
    def get_remote_info(cls, identifier: int) -> Self:
        """
        A UDP message to get information about the remote system.

        :param identifier: the identifier to use for this UDP message
        :returns: an AdsUDPMessage instance
        """
        blank_response = AdsUDPResponseStream(
            netid=AmsNetId.from_string("0.0.0.0.0.0").to_bytes(),
            port=0,
            count=0,
            data=b"",
        )  # not required to be valid, just a response template

        return cls(
            udp_cookie=0x71146603,  # b"\x71\x14\x66\x03",
            invoke_id=identifier,
            service_id=SystemServiceCommandId.ADSSVCID_READSERVICEINFO,
            data=blank_response.to_bytes(),
        )

    @classmethod
    def add_remote_route(cls, identifier: int, route_info: bytes) -> Self:
        """
        A UDP message to add a remote route to the system.

        :param identifier: the identifier to use for this UDP message
        :param route_info: the route information data in bytes
        :returns: an AdsUDPMessage instance
        """
        return cls(
            udp_cookie=0x71146603,  # b"\x71\x14\x66\x03",
            invoke_id=identifier,
            service_id=SystemServiceCommandId.ADSSVCID_ADDROUTE,
            data=route_info,
        )

    @classmethod
    def del_remote_route(cls, identifier: int, route_info: bytes) -> Self:
        """
        A UDP message to delete a remote route from the system.

        :param identifier: the identifier to use for this UDP message
        :param route_info: the route information data in bytes
        :returns: an AdsUDPMessage instance
        """
        return cls(
            udp_cookie=0x71146603,  # b"\x71\x14\x66\x03",
            invoke_id=identifier,
            service_id=SystemServiceCommandId.ADSSVCID_DELROUTE,
            data=route_info,
        )


class AdsUDPResponseStream(Message):
    """Define the ADS UDP response stream structure."""

    netid: BYTES6
    """The Ams NetID of the remote target"""
    port: UINT16
    """The Ams Port of the remote target"""
    count: UINT32
    """The number of UDP messages in the stream"""
    data: bytes
    """The UDP response stream data"""


# ===================================================================
# ===== TCP COMMUNICATIONS
# ===================================================================


class AmsHeader(Message):
    """
    AMS Header structure included in all ADS communications.
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/115847307.html&id=7738940192708835096
    """

    target_net_id: BYTES6
    """The AMS netid of the station for which the packet is intended"""
    target_port: UINT16
    """The AMS port of the station for which the packet is intended"""
    source_net_id: BYTES6
    """The AMS netid of the station from which the packet is sent"""
    source_port: UINT16
    """The AMS port of the station from which the packet is sent"""
    command_id: CommandId
    """ADS command id"""
    state_flags: StateFlag
    """Defines the protocol (bit7: TCP/UDP), interface (bit3: ADS) and message type \
        (bit1: request/response)"""
    length: UINT32
    """Length of the data in bytes attached to this header"""
    error_code: ErrorCode
    """ADS error number"""
    invoke_id: UINT32
    """Id used to map a received response to a sent request"""


# ===================================================================
# ===== ETHERCAT I/O PROPERTIES
# ===================================================================


class IOIdentity(Message):
    """
    Define the identity parameters of an EtherCAT device or slave.
    """

    vendor_id: UINT32
    """The vendor id number"""
    product_code: UINT32
    """The product code"""
    revision_number: UINT32
    """The revision number"""
    serial_number: UINT32
    """The serial number"""

    def __str__(self):
        return f"code:{self.product_code}, s/n:{self.serial_number}"


class DeviceFrames(Message):
    """
    Define the frame counters of an EtherCAT device.
    """

    time: UINT32
    """System time"""
    cyclic_sent: UINT32
    """Number of cyclic frames sent by the master device"""
    cyclic_lost: UINT32
    """Number of lost cyclic frames"""
    acyclic_sent: UINT32
    """Number of acyclic frames sent by the master device"""
    acyclic_lost: UINT32
    """Number of lost acyclic frames"""

    def __array__(self):
        return np.array(
            [
                self.time,
                self.cyclic_sent,
                self.cyclic_lost,
                self.acyclic_sent,
                self.acyclic_lost,
            ],
            dtype=np.uint32,
        )

    @property
    def __array_struct__(self):
        return self.__array__().__array_struct__


class SlaveCRC(Message):
    """
    Define the cyclic redundancy check error counters of an EtherCAT slave.
    Ports B, C and D may not be used, thus potentially absent from an ADS response.
    """

    port_a_crc: UINT32
    """CRC error counter of communication port A"""
    port_b_crc: UINT32
    """CRC error counter of communication port B"""
    port_c_crc: UINT32
    """CRC error counter of communication port C"""
    port_d_crc: UINT32
    """CRC error counter of communication port D"""

    def __array__(self):
        return np.array(
            [
                self.portA_crc,
                self.portB_crc,
                self.portC_crc,
                self.portD_crc,
            ],
            dtype=np.uint32,
        )

    @property
    def __array_struct__(self):
        return self.__array__().__array_struct__


class SlaveState(Message):
    """
    Define the EtherCAT state and link status of an EtherCAT slave.
    """

    ecat_state: UINT8
    """The EtherCAT state"""
    link_status: UINT8
    """The link status for communication"""

    def __array__(self):
        return np.array(
            [
                self.ecat_state,
                self.link_status,
            ],
            dtype=np.uint8,
        )

    @property
    def __array_struct__(self):
        return self.__array__().__array_struct__


class AdsSymbolTableInfo(Message):
    """
    Define the symbol table information stored on the I/O server.
    """

    symbol_count: UINT32
    """The number of symbols accessible on the device"""
    table_length: UINT32
    """The length of the symbol table uploaded in the controller"""
    reserved: BYTES12
    """Unidentified additional data used by TwinCAT"""


class AdsSymbolTableEntry(Message):
    """
    ADS devices that support symbol names store those names in an internal table.
    """

    read_length: UINT32
    """Length of the complete symbol entry data in bytes"""
    index_group: UINT32
    """Index group of the symbol"""
    index_offset: UINT32
    """Index offset of the symbol"""
    size: UINT32
    """Size of the symbol in bytes (0 corresponds to 'bit')"""
    ads_type: AdsDataType
    """ADS data type of the symbol"""
    flag: SymbolFlag
    """ADS symbol flag"""
    name_size: UINT16
    """Length of the symbol name in bytes (null terminating character not counted)"""
    type_size: UINT16
    """Length of the symbol type name in bytes \
        (null terminating character not counted)"""
    comment_size: UINT16
    """Length of the entry comment in bytes (null terminating character not counted)"""
    data: bytes
    """Symbol entry data"""


# ===================================================================
# ===== INFO
# ===================================================================


class AdsReadDeviceInfoRequest(MessageRequest):
    """
    ADS Read device Info packet
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/115875851.html&id=8158832529229503828
    """

    pass  # No additional data required


class AdsReadDeviceInfoResponse(MessageResponse):
    """
    ADS Read Device Info data structure received in response to an \
        ADS Read Device Info request.
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/115875851.html&id=8158832529229503828
    """

    result: ErrorCode
    """ADS error number"""
    major_version: UINT8
    """Major version number of the ADS device"""
    minor_version: UINT8
    """Minor version number of the ADS device"""
    version_build: UINT16
    """Build number"""
    device_name: BYTES16
    """Name of the ADS device"""


# ===================================================================
# ===== STATE
# ===================================================================


class AdsReadStateRequest(MessageRequest):
    """
    ADS Read State packet
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/115878923.html&id=6874981934243835072
    """

    pass  # No additional data required


class AdsReadStateResponse(MessageResponse):
    """
    ADS Read State data structure received in response to an ADS Read State request.
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/115878923.html&id=6874981934243835072
    """

    result: ErrorCode
    """ADS error number"""
    ads_state: UINT16
    """ADS status"""
    device_state: UINT16
    """Device status"""


# ===================================================================
# ===== READ
# ===================================================================


class AdsReadRequest(MessageRequest):
    """
    ADS Read packet
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/115876875.html&id=4960931295000833536
    """

    index_group: IndexGroup
    """Index group of the data"""
    index_offset: UINT32
    """Index offset of the data"""
    read_length: UINT32
    """Length of the data in bytes which is read"""

    @classmethod
    def read_device_count(cls) -> Self:
        """
        An ADS request to read the number of devices registered with the I/O server.

        :returns: an AdsReadRequest message
        """
        return cls(
            index_group=IndexGroup.ADSIGR_IODEVICE_STATE_BASE,
            index_offset=0x2,
            read_length=np.dtype(np.uint32).itemsize,
        )

    @classmethod
    def read_device_ids(cls, device_count: SupportsInt) -> Self:
        """
        An ADS request to read the id of the devices registered with the I/O server.
        (Note: the first index will represent the device count; device ids will follow)

        :param device_count: the number of registered EtherCAT devices

        :returns: an AdsReadRequest message
        """
        return cls(
            index_group=IndexGroup.ADSIGR_IODEVICE_STATE_BASE,
            index_offset=0x1,
            read_length=(int(device_count) + 1) * np.dtype(np.uint16).itemsize,
        )

    @classmethod
    def read_device_type(cls, device_id: int) -> Self:
        """
        An ADS request to read the type of a given EtherCAT device.

        :param device_id: the id of the device

        :returns: an AdsReadRequest message
        """
        return cls(
            index_group=IndexGroup(
                int(IndexGroup.ADSIGR_IODEVICE_STATE_BASE + device_id)
            ),
            index_offset=0x7,
            read_length=np.dtype(np.uint16).itemsize,
        )

    @classmethod
    def read_device_name(cls, device_id: int) -> Self:
        """
        An ADS request to read the name of a given EtherCAT device.

        :param device_id: the id of the device

        :returns: an AdsReadRequest message
        """
        return cls(
            index_group=IndexGroup(
                int(IndexGroup.ADSIGR_IODEVICE_STATE_BASE + device_id)
            ),
            index_offset=0x1,
            read_length=0xFF,
        )

    @classmethod
    def read_device_netid(cls, device_id: int) -> Self:
        """
        An ADS request to read the ams netid of a given EtherCAT device.

        :param device_id: the id of the device

        :returns: an AdsReadRequest message
        """
        return cls(
            index_group=IndexGroup(
                int(IndexGroup.ADSIGR_IODEVICE_STATE_BASE + device_id)
            ),
            index_offset=0x5,
            read_length=6,
        )

    @classmethod
    def read_device_identity(cls, subindex: str) -> Self:
        """
        An ADS request to read the CANopen identity of a given EtherCAT device
        (this includes vendorId, productCode, revisionNumber and serialNumber).
        The value is accessed via a CAN-over-EtherCAT parameter (sdo).

        :returns: an AdsReadRequest message
        """
        index = "0x1018"
        return cls(
            index_group=IndexGroup.ADSIGRP_COE_LINK,
            index_offset=int(index + subindex, base=16),
            read_length=np.dtype(np.uint32).itemsize,
        )

    @classmethod
    def read_slave_count(cls) -> Self:
        """
        An ADS request to read the number of slave terminals configured on a device.

        :returns: an AdsReadRequest message
        """
        return cls(
            index_group=IndexGroup.ADSIGRP_MASTER_COUNT_SLAVE,
            index_offset=0x0,
            read_length=np.dtype(np.uint16).itemsize,
        )

    @classmethod
    def read_slaves_addresses(cls, num_slaves: int) -> Self:
        """
        An ADS request to read the EtherCAT addresses of all configured slave terminals.

        :param num_slaves: the number of slave terminals on the EtherCAT device

        :returns: an AdsReadRequest message
        """
        return cls(
            index_group=IndexGroup.ADSIGRP_MASTER_SLAVE_ADDRESSES,
            index_offset=0x0,
            read_length=(np.dtype(np.uint16).itemsize) * num_slaves,
        )

    @classmethod
    def read_slave_identity(cls, address: SupportsInt) -> Self:
        """
        An ADS request to read the CANopen identity of a configured slave terminal.
        (this includes vendorId, productCode, revisionNumber and serialNumber)

        :param address: the address of the slave terminal on the EtherCAT device

        :returns: an AdsReadRequest message
        """
        return cls(
            index_group=IndexGroup.ADSIGRP_MASTER_SLAVE_IDENTITY,
            index_offset=np.uint32(address),
            read_length=(np.dtype(np.uint32).itemsize) * 4,
        )

    @classmethod
    def read_device_state(cls) -> Self:
        """
        An ADS request to read the state of an EtherCAT device (e.g. Master device).

        :returns: an AdsReadRequest message
        """
        return cls(
            index_group=IndexGroup.ADSIGRP_MASTER_STATEMACHINE,
            index_offset=0x0100,
            read_length=np.dtype(np.uint16).itemsize,
        )

    @classmethod
    def read_slaves_states(cls, num_slaves: SupportsInt) -> Self:
        """
        An ADS request to read the EtherCAT state and link status of all slave terminal.

        :param num_slaves: the number of slave terminals on the EtherCAT device

        :returns: an AdsReadRequest message
        """
        return cls(
            index_group=IndexGroup.ADSIGRP_SLAVE_STATEMACHINE,
            index_offset=0x0,
            read_length=(np.dtype(np.uint16).itemsize) * int(num_slaves),
        )

    @classmethod
    def read_slave_states(cls, address: SupportsInt) -> Self:
        """
        An ADS request to read the EtherCAT state and link status of \
            a single slave terminal.

        :param address: the address of the slave terminal on the EtherCAT device

        :returns: an AdsReadRequest message
        """
        return cls(
            index_group=IndexGroup.ADSIGRP_SLAVE_STATEMACHINE,
            index_offset=np.uint32(address),
            read_length=np.dtype(np.uint16).itemsize,
        )

    @classmethod
    def read_slaves_crc(cls, num_slaves: SupportsInt) -> Self:
        """
        An ADS request to read the counter values sum for the cyclic redundancy check \
            (CRC) of all slaves.
        CRC counters are incremented for the respective communication ports (A,B,C,D) \
            if an error has occurred (e.g. frames passing through the network which \
                are destroyed or damaged due to cable, contact or connector problems).

        :param num_slaves: the number of slave terminals on the device

        :returns: an AdsReadRequest message
        """
        return cls(
            index_group=IndexGroup.ADSIGRP_SLAVE_CRC_COUNTERS,
            index_offset=0x0,
            read_length=(np.dtype(np.uint32).itemsize) * int(num_slaves),
        )

    @classmethod
    def read_slave_crc(cls, address: SupportsInt) -> Self:
        """
        An ADS request to read the counter values for the cyclic redundancy check \
            (CRC) of a single slave terminal.
        CRC counters are incremented for the respective communication ports (A,B,C,D) \
            if an error has occurred (e.g. frames passing through the network which \
                are destroyed or damaged due to cable, contact or connector problems).

        :param address: the address of the slave terminal on the EtherCAT device

        :returns: an AdsReadRequest message
        """
        return cls(
            index_group=IndexGroup.ADSIGRP_SLAVE_CRC_COUNTERS,
            index_offset=np.uint32(address),
            read_length=(np.dtype(np.uint32).itemsize) * 4,
        )

    @classmethod
    def read_device_frame_counters(cls) -> Self:
        """
        An ADS request to read the frame counters of an EtherCAT device.
        This includes cyclic and acyclic frames, both sent and lost.

        :returns: an AdsReadRequest message
        """
        return cls(
            index_group=IndexGroup.ADSIGRP_MASTER_FRAME_COUNTERS,
            index_offset=0x0,
            read_length=(np.dtype(np.uint32).itemsize) * 5,
        )

    @classmethod
    def read_slave_type(cls, index: str) -> Self:
        """
        An ADS request to read the type of a slave terminal configured on a device.
        The value is accessed via a CAN-over-EtherCAT parameter (sdo).

        :param index: the index of the accessed CoE range as an hexadecimal string

        :returns: an AdsReadRequest message
        """
        subindex = "0002"
        return cls(
            index_group=IndexGroup.ADSIGRP_COE_LINK,
            index_offset=int(index + subindex, base=16),
            read_length=16,
        )

    @classmethod
    def read_slave_name(cls, index: str) -> Self:
        """
        An ADS request to read the name of a slave terminal configured on a device.
        The value is accessed via a CAN-over-EtherCAT parameter (sdo).

        :param index: the index of the accessed CoE range as an hexadecimal string

        :returns: an AdsReadRequest message
        """
        subindex = "0003"
        return cls(
            index_group=IndexGroup.ADSIGRP_COE_LINK,
            index_offset=int(index + subindex, base=16),
            read_length=32,
        )

    @classmethod
    def get_length_symbol_table(cls) -> Self:
        """
        An ADS request to get the length of the symbol table registered with the server.

        :returns: an AdsReadRequest message
        """
        return cls(
            index_group=IndexGroup.ADSIGRP_SYM_UPLOADINFO2,
            index_offset=0,
            read_length=24,
        )

    @classmethod
    def fetch_symbol_table(cls, length: SupportsInt) -> Self:
        """
        ADS request to get a list of all symbols defined in the server's symbol table.

        :param length: the length of the symbol table data

        :returns: an AdsReadRequest message
        """
        return cls(
            index_group=IndexGroup.ADSIGRP_SYM_UPLOAD,
            index_offset=0,
            read_length=length,
        )

    @classmethod
    def read_symbol(
        cls, group: SupportsInt, offset: SupportsInt, length: SupportsInt
    ) -> Self:
        """
        An ADS request to get the value of a given ADS symbol.

        :param group: the index group of the ADS symbol to read the value from
        :param offset: the index offset of the ADS symbol to read the value from
        :param length: the length of the symbol data in bytes to read

        :returns: an AdsReadRequest message
        """
        return cls(
            index_group=IndexGroup(group),
            index_offset=offset,
            read_length=length,
        )

    @classmethod
    def read_coe_value(cls, index: str, subindex: str, datatype: Any) -> Self:
        """
        An ADS request to get the value of a CAN-over-EtherCAT parameter (sdo/pdo).

        :param index: the CoE index assigned to the parameter (HIWORD=0xYYYY0000)
        :param subindex: the CoE subindex assigned to the parameter (LOBYTE=0x000000YY)
        :param type: the parameter data type

        :returns: an AdsReadRequest message
        """
        return cls(
            index_group=IndexGroup.ADSIGRP_COE_LINK,
            index_offset=int(index + subindex, base=16),
            read_length=np.dtype(datatype).itemsize,
        )


class AdsReadResponse(MessageResponse):
    """
    ADS Read data structure received in response to an ADS Read request.
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/115876875.html&id=4960931295000833536
    """

    result: ErrorCode
    """ADS error number"""
    length: UINT32
    """Length of the data supplied back from the ADS device"""
    data: bytes
    """Data supplied back from the ADS device"""


# ===================================================================
# ===== WRITE
# ===================================================================


class AdsWriteRequest(MessageRequest):
    """
    ADS Write packet
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/115877899.html&id=8845698684103663373
    """

    index_group: IndexGroup
    """Index group of the data"""
    index_offset: UINT32
    """Index offset of the data"""
    write_length: UINT32
    """Length of the data in bytes which is written"""
    data: bytes
    """Data written to the ADS device"""

    @classmethod
    def reset_device_frame_counters(cls) -> Self:
        """
        An ADS request to reset the frame counters of an EtherCAT device to zero.

        :returns: an AdsReadRequest message
        """
        return cls(
            index_group=IndexGroup.ADSIGRP_MASTER_FRAME_COUNTERS,
            index_offset=0x0,
            write_length=0x0,
            data=b"",
        )

    @classmethod
    def write_symbol(cls, group: SupportsInt, offset: SupportsInt, data: bytes) -> Self:
        """
        An ADS request to write a given value to an ADS symbol.

        :param group: the index group of the ADS symbol to write the value to
        :param offset: the index offset of the ADS symbol to write the value to
        :param data: the data in bytes to write

        :returns: an AdsWriteRequest message
        """
        return cls(
            index_group=IndexGroup(group),
            index_offset=offset,
            write_length=len(data),
            data=data,
        )

    @classmethod
    def write_coe_value(cls, index: str, subindex: str, data: bytes) -> Self:
        """
        An ADS request to write a given value to a CAN-over-EtherCAT parameter
        (sdo/pdo).

        :param index: the CoE index assigned to the parameter (HIWORD=0xYYYY0000)
        :param subindex: the CoE subindex assigned to the parameter (LOBYTE=0x000000YY)
        :param data: the data in bytes to write

        :returns: an AdsWriteRequest message
        """
        return cls(
            index_group=IndexGroup.ADSIGRP_COE_LINK,
            index_offset=int(index + subindex, base=16),
            write_length=len(data),
            data=data,
        )


class AdsWriteResponse(MessageResponse):
    """
    ADS Write data structure received in response to an ADS Write request.
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/115877899.html&id=8845698684103663373
    """

    result: ErrorCode
    """ADS error number"""


# ===================================================================
# ===== READ/WRITE
# ===================================================================


class AdsReadWriteRequest(MessageRequest):
    """
    ADS Read Write packet
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/115884043.html&id=2085949217954035635
    """

    index_group: IndexGroup
    """Index group of the data"""
    index_offset: UINT32
    """Index offset of the data"""
    read_length: UINT32
    """Length of the data in bytes which is read"""
    write_length: UINT32
    """Length of the data in bytes which is written"""
    data: bytes
    """Data written to the ADS device"""

    @classmethod
    def get_handle_by_name(cls, name: str) -> Self:
        """
        An ADS request to get a unique handle associated with a given name.

        :param name: the ads symbol variable name

        :returns: an AdsReadWriteRequest message
        """
        data = name.encode(encoding=TWINCAT_STRING_ENCODING) + b"\x00"
        return cls(
            index_group=IndexGroup.ADSIGR_GET_SYMHANDLE_BYNAME,
            index_offset=0,
            read_length=np.dtype(np.uint32).itemsize,
            write_length=len(data),
            data=data,
        )

    @classmethod
    def readwrite_symbol(
        cls, group: SupportsInt, offset: SupportsInt, length: SupportsInt, data: bytes
    ) -> Self:
        """
        An ADS request to read the current value of a symbol \
            and write a new value to it as part of a single data frame.

        :param group: the index group of the ADS symbol to read from and write to
        :param offset: the index offset of the ADS symbol to read from and write to
        :param length: the length of the symbol data in bytes to read
        :param data: the data in bytes to write

        :returns: an AdsReadWriteRequest message
        """
        return cls(
            index_group=IndexGroup(group),
            index_offset=offset,
            read_length=length,
            write_length=len(data),
            data=data,
        )

    @classmethod
    def sumread_symbols(cls, reads: Sequence[AdsReadRequest]) -> Self:
        """
        Get the value of multiple variables which are not structured \
            within a linear memory.
        This function appears to be valid for ADS symbol variables, but not applicable \
            to CoE parameters (returns invalid index group error).

        :param reads: a list of AdsReadRequest sub-commands, \
            each one of them associated with one of the variables to read

        :returns: an AdsReadWriteRequest message
        """
        count = len(reads)
        assert count, (
            "Minimum one ADS ReadRequest is required to carry out an ADS SumRead call"
        )
        assert count < 500, (
            "A maximum of 500 sub-commands can be requested in an ADS SumRead call."
        )
        data = b""
        length = count * np.dtype(ErrorCode).itemsize
        for request in reads:
            data += request.to_bytes(include_data=False)
            length += request.read_length

        return cls(
            index_group=IndexGroup.ADSIGRP_SUMUP_READ,
            index_offset=count,
            read_length=length,
            write_length=len(data),
            data=data,
        )

    @classmethod
    def sumwrite_symbols(cls, writes: Sequence[AdsWriteRequest]) -> Self:
        """
        Set the value of multiple variables which are not structured \
            within a linear memory.
        This function appears to be valid for ADS symbol variables, but not applicable \
            to CoE parameters (returns invalid index group error).

        !Note that this service appears not to be supported by the CX2020 server \
            (error 0x0701 ADSERR_DEVICE_SRVNOTSUPP)!

        :param writes: a list of AdsWriteRequest sub-commands, each one of them \
            associated with one of the variables to write

        :returns: an AdsReadWriteRequest message
        """
        count = len(writes)
        assert count, (
            "Minimum one ADS WriteRequest is required to carry out an ADS SumWrite call"
        )
        assert count < 500, (
            "A maximum of 500 sub-commands can be requested in an ADS SumWrite call."
        )
        data = b""
        length = count * np.dtype(ErrorCode).itemsize
        for request in writes:
            data += request.to_bytes(include_data=False)
        for request in writes:
            data += request.data

        return cls(
            index_group=IndexGroup.ADSIGRP_SUMUP_WRITE,
            index_offset=count,
            read_length=length,
            write_length=len(data),
            data=data,
        )

    @classmethod
    def sumreadwrite_symbols(cls, readwrites: Sequence[Self]) -> Self:
        """
        Get and set the value of multiple variables which are not structured \
            within a linear memory.
        This function appears to be valid for ADS symbol variables, but not applicable \
            to CoE parameters (returns invalid index group error).

        !Note that this service appears not to be supported by the CX2020 server \
            (error 0x0701 ADSERR_DEVICE_SRVNOTSUPP)!

        :param readwrites: a list of AdsReadWriteRequest sub-commands, \
            each one of them associated with one of the variables to read then write

        :returns: an AdsReadWriteRequest message
        """
        count = len(readwrites)
        assert count, (
            "Minimum one ADS ReadWrite request is required to carry out \
                an ADS SumReadWrite call"
        )
        assert count < 500, (
            "A maximum of 500 sub-commands can be requested in \
                an ADS SumReadWrite call."
        )
        data = b""
        length = count * (np.dtype(ErrorCode).itemsize + np.dtype(np.uint32).itemsize)
        for cmd in readwrites:
            length += cmd.read_length
            data += cmd.to_bytes(include_data=False)
        for cmd in readwrites:
            data += cmd.data

        return cls(
            index_group=IndexGroup.ADSIGRP_SUMUP_READWRITE,
            index_offset=count,
            read_length=length,
            write_length=len(data),
            data=data,
        )


class AdsReadWriteResponse(MessageResponse):
    """
    ADS ReadWrite data structure received in response to an ADS ReadWrite request.
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/115884043.html&id=2085949217954035635
    """

    result: ErrorCode
    """ADS error number"""
    length: UINT32
    """Length of the data supplied back from the ADS device"""
    data: bytes
    """Data supplied back from the ADS device"""


# ===================================================================
# ===== ADD DEVICE NOTIFICATION
# ===================================================================


class AdsAddDeviceNotificationRequest(MessageRequest):
    """
    ADS Add Device Notification packet
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/115880971.html&id=7388557527878561663
    """

    index_group: IndexGroup
    """Index group of the data"""
    index_offset: UINT32
    """Index offset of the data"""
    length: UINT32
    """Length of data in bytes expected for this notification"""
    transmission_mode: TransmissionMode = TransmissionMode.ADSTRANS_SERVERONCHA
    """Chosen mode of device notification"""
    max_delay: UINT32
    """At the latest after this time, the ADS Device Notification is called; \
        the unit is 100ns"""
    cycle_time: UINT32
    """The ADS server checks if the value changes in this time slice; \
        the unit is 100ns"""
    reserved: BYTES16 = BYTES16(default=b"")
    """Reserved memory block,; must be set to 0"""


class AdsAddDeviceNotificationResponse(MessageResponse):
    """
    ADS AddDeviceNotification data structure received in response to an ADS \
        AddDeviceNotification request.
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/115880971.html&id=7388557527878561663
    """

    result: ErrorCode
    """ADS error number"""
    handle: UINT32
    """Notification handle"""


# ===================================================================
# ===== DELETE DEVICE NOTIFICATION
# ===================================================================


class AdsDeleteDeviceNotificationRequest(MessageRequest):
    """
    ADS Delete Device Notification packet
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/115881995.html&id=6216061301016726131
    """

    handle: UINT32
    """Notification handle; this handle is created by the AdsAddDeviceNotification \
        command"""


class AdsDeleteDeviceNotificationResponse(MessageResponse):
    """
    ADS DeleteDeviceNotification data structure received in response to an ADS \
        DeleteDeviceNotification request.
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/115881995.html&id=6216061301016726131
    """

    result: ErrorCode
    """ADS error number"""


# ===================================================================
# ===== DEVICE NOTIFICATION
# ===================================================================


class AdsCombinedNotificationStream(Message):
    """
    Define a combined stream made from multiple notification streams.
    Such object is required for very high number of symbol notifications; \
        each ads notification stream can comprise around 4kbytes.
    """

    size: UINT16
    """Number of AdsNotificationStream elements"""
    data: bytes
    """Array of AdsNotificationStream elements"""

    def get_combined_notifications_dtype(
        self,
        device_name: str,
        symbols: dict[SupportsInt, Any],
    ) -> np.dtype:
        """
        Get the datatype structure used to interpret the bytes array from a \
            notification which comprises multiple streams

        :param device_name: the name of the device providing the notification
        :param symbols: a container object mapping the notification handle and related \
            symbol for all configured notifications

        :returns: a numpy array of datatypes characteristic of a whole ads notification
        """
        dtypes = []
        offset = 0
        stream_size = 0

        data = self.data
        for i in range(1, self.size + 1):
            stream_size = 4 + int.from_bytes(data[:4], byteorder="little", signed=False)
            notification = AdsNotificationStream.from_bytes(data[:stream_size])
            dtypes += [
                (f"_{device_name}.length{i}", np.uint32),
                (f"_{device_name}.stamps{i}", np.uint32),
            ]
            assert notification.stamps == 1, (
                f"Error: notification stream {i} comprises {notification.stamps} \
                    distinct timestamps."
            )
            stamp_header = AdsStampHeader.from_bytes(notification.data)
            dtypes += [
                (f"_{device_name}.timestamp{i}", np.uint64),
                (f"_{device_name}.samples{i}", np.uint32),
            ]
            notif_data = stamp_header.data

            for _ in range(int(stamp_header.samples)):
                assert notif_data, notif_data
                sample = AdsNotificationSample.from_bytes(notif_data)
                symbol = symbols[sample.handle]
                assert symbol.nbytes == sample.size
                dtypes += [
                    (f"_{symbol.name.replace(' ', '')}.handle", np.uint32),
                    (f"_{symbol.name.replace(' ', '')}.size", np.uint32),
                    (f"_{symbol.name.replace(' ', '')}.value", symbol.datatype),
                ]
                notif_data = notif_data[8 + sample.size :]

            assert notif_data == b"", (
                f"Error: unprocessed data in the notification stream {i}: {notif_data}"
            )
            offset += stream_size
            data = self.data[offset:]

        assert data == b"", (
            f"Error: unprocessed data in the combined notification stream: {data}"
        )

        return np.dtype(dtypes)


class AdsNotificationStream(Message):
    """
    ADS DeviceNotification packet
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/115883019.html&id=3360461216738457777
    """

    length: UINT32
    """Size of 'stamps+data' in bytes"""
    stamps: UINT32
    """Number of AdsStampHeader elements"""
    data: bytes
    """Array with AdsStampHeader elements"""

    def get_notification_dtype(
        self, device_name: str, symbols: dict[SupportsInt, Any]
    ) -> np.dtype:
        """
        Get the datatype structure used to interpret the bytes array from a single \
            notification packet (note that the notification packet is assumed to \
                comprise a single stamp header).

        :param device_name: the name of the device providing the notification
        :param symbols: a container object mapping the notification handle and related \
            symbol for all configured notifications

        :returns: a numpy array of datatypes characteristic of the active device \
            notification structure
        """
        dtypes = [
            (f"_{device_name}.length", np.uint32),
            (f"_{device_name}.stamps", np.uint32),
        ]
        assert self.stamps == 1, (
            f"Error: notification comprises {self.stamps} distinct timestamps."
        )
        stamp_header = AdsStampHeader.from_bytes(self.data)
        dtypes += [
            (f"_{device_name}.timestamp", np.uint64),
            (f"_{device_name}.samples", np.uint32),
        ]
        data = stamp_header.data
        for _ in range(int(stamp_header.samples)):
            assert data, data
            sample = AdsNotificationSample.from_bytes(data)
            symbol = symbols[sample.handle]
            assert symbol.nbytes == sample.size
            dtypes += [
                (f"_{symbol.name.replace(' ', '')}.handle", np.uint32),
                (f"_{symbol.name.replace(' ', '')}.size", np.uint32),
                (f"_{symbol.name.replace(' ', '')}.value", symbol.datatype),
            ]
            data = data[8 + sample.size :]
        assert data == b"", (
            f"Error: unprocessed data in the notification stream: {data}"
        )
        return np.dtype(dtypes)


class AdsStampHeader(Message):
    """
    ADS Stamp Header data structure.
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/115883019.html&id=3360461216738457777"""

    timestamp: UINT64
    """UTC time stamp corresponding to the notification packet \
        (100ns intervals since 01.01.1601)"""
    samples: UINT32
    """Number of AdsNotificationSample elements"""
    data: bytes
    """Array with AdsNotificationSample elements"""


class AdsNotificationSample(Message):
    """
    ADS Notification Sample data structure.
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/115883019.html&id=3360461216738457777
    """

    handle: UINT32
    """Notification handle"""
    size: UINT32
    """Size of the data in bytes"""
    data: bytes
    """Data"""


# ===================================================================
# ===== MESSAGE MAPPING
# ===================================================================

# Dictionary of all available ADS messages.
MESSAGE_CLASS: dict[type[MessageRequest], type[MessageResponse]] = {
    AdsAddDeviceNotificationRequest: AdsAddDeviceNotificationResponse,
    AdsDeleteDeviceNotificationRequest: AdsDeleteDeviceNotificationResponse,
    AdsReadDeviceInfoRequest: AdsReadDeviceInfoResponse,
    AdsReadStateRequest: AdsReadStateResponse,
    AdsReadRequest: AdsReadResponse,
    AdsReadWriteRequest: AdsReadWriteResponse,
    AdsWriteRequest: AdsWriteResponse,
}

# Dictionary of all available ADS requests and associated commands.
REQUEST_CLASS: dict[type[MessageRequest], CommandId] = {
    AdsAddDeviceNotificationRequest: CommandId.ADSSRVID_ADDDEVICENOTE,
    AdsDeleteDeviceNotificationRequest: CommandId.ADSSRVID_DELETEDEVICENOTE,
    AdsReadDeviceInfoRequest: CommandId.ADSSRVID_READDEVICEINFO,
    AdsReadStateRequest: CommandId.ADSSRVID_READSTATE,
    AdsReadRequest: CommandId.ADSSRVID_READ,
    AdsReadWriteRequest: CommandId.ADSSRVID_READWRITE,
    AdsWriteRequest: CommandId.ADSSRVID_WRITE,
}

# Dictionary of all available ADS commands and associated responses.
RESPONSE_CLASS: dict[CommandId, type[MessageResponse]] = {
    CommandId.ADSSRVID_ADDDEVICENOTE: AdsAddDeviceNotificationResponse,
    CommandId.ADSSRVID_DELETEDEVICENOTE: AdsDeleteDeviceNotificationResponse,
    CommandId.ADSSRVID_READDEVICEINFO: AdsReadDeviceInfoResponse,
    CommandId.ADSSRVID_READSTATE: AdsReadStateResponse,
    CommandId.ADSSRVID_READ: AdsReadResponse,
    CommandId.ADSSRVID_READWRITE: AdsReadWriteResponse,
    CommandId.ADSSRVID_WRITE: AdsWriteResponse,
}



"""
CATio Performance Tests file for the CATio system.
Each test prints detailed analysis metrics (ops/sec, latency in ms, comparative ratios).
The tests use the same mock server infrastructure and port to avoid conflicts.

Tests currently measure and validate performance characteristics for:
- Operation latency (read, write, notifications)
- Throughput (operations per second)
- Scalability with concurrent operations
- Memory usage and leaks
- Load and stress testing
- Comparative operation modes
- Bottleneck identification

Run the tests with:
```bash
python -m pytest tests/test_catio_performance.py -v
```
"""

import asyncio
import random
import time
from typing import Any

import numpy as np
import pytest

from fastcs_catio.client import AsyncioADSClient
from fastcs_catio.devices import AdsSymbol
from mock_server import MockADSServer

IO_SERVER_PORT: int = 25565

pytest.skip(allow_module_level=True, reason="TODO these are all failing")


@pytest.fixture
async def perf_test_server():
    """Set up a mock ADS server for performance testing."""
    server = MockADSServer(host="127.0.0.1", port=48898)

    # Pre-populate the mock server with test symbol data
    for i in range(200):
        server.set_symbol_data(f"Symbol{i}", np.int32(i))

    await server.start()
    yield server
    await server.stop()


# ===================================================================
# Latency Measurement Tests
# ===================================================================


class TestOperationLatency:
    """Test and measure operation latencies."""

    @pytest.mark.asyncio
    async def test_read_state_latency(self, perf_test_server: MockADSServer):
        """Measure read operation latency for the mock server states."""
        # Connect to the mock server
        client = await AsyncioADSClient.connected_to(
            target_ip=perf_test_server.host,
            target_ams_net_id=perf_test_server.local_netid.to_string(),
            target_ams_port=IO_SERVER_PORT,
            ads_port=perf_test_server.port,
        )

        # Define the test sample population
        num_operations = 100

        try:
            # First warmup call to confirm functionality
            await client._get_ioserver_states()

            # Latency measurements
            latencies = []
            for _ in range(num_operations):
                start = time.perf_counter()
                await client._get_ioserver_states()
                latency = time.perf_counter() - start
                latencies.append(latency)

            # Calculate statistics
            avg_latency = sum(latencies) / len(latencies)
            min_latency = min(latencies)
            max_latency = max(latencies)

            # Verify reasonable latencies (should be sub-second for mock)
            assert avg_latency < 1.0
            assert min_latency < max_latency

            # Report results to stdout
            print(
                f"Read states -- latency test results over {num_operations} samples:\n"
                + f"Avg: {avg_latency * 1000:.2f}ms,\n"
                + f"Min: {min_latency * 1000:.2f}ms,\n"
                + f"Max: {max_latency * 1000:.2f}ms"
            )

        finally:
            # Disconnect the client
            await client.close()

    @pytest.mark.asyncio
    async def test_read_latency(self, perf_test_server: MockADSServer):
        """Measure symbol read operation latency."""
        # Connect to the mock server
        client = await AsyncioADSClient.connected_to(
            target_ip=perf_test_server.host,
            target_ams_net_id=perf_test_server.local_netid.to_string(),
            target_ams_port=IO_SERVER_PORT,
            ads_port=perf_test_server.port,
        )

        # Define the test sample population
        num_operations = 100

        # Create a sequence of random symbols for testing
        test_symbols: list[AdsSymbol] = []
        for _ in range(num_operations):
            test_symbols.append(perf_test_server.get_random_symbol())
        assert len(test_symbols) == num_operations

        try:
            # First warmup call to confirm functionality
            await client.read_ads_symbol(test_symbols[0])

            # Latency measurements
            latencies = []
            for symbol in test_symbols:
                start = time.perf_counter()
                await client.read_ads_symbol(symbol)
                latency = time.perf_counter() - start
                latencies.append(latency)

            # Calculate statistics
            avg_latency = sum(latencies) / len(latencies)
            min_latency = min(latencies)
            max_latency = max(latencies)

            # Verify reasonable latencies (should be sub-second for mock)
            assert avg_latency < 1.0
            assert min_latency < max_latency

            # Report results to stdout
            print(
                f"Read symbol -- latency test results over {num_operations} samples:\n"
                + f"Avg: {avg_latency * 1000:.2f}ms,\n"
                + f"Min: {min_latency * 1000:.2f}ms,\n"
                + f"Max: {max_latency * 1000:.2f}ms"
            )

        finally:
            # Disconnect the client
            await client.close()

    @pytest.mark.asyncio
    async def test_write_latency(self, perf_test_server: MockADSServer):
        """Measure symbol write operation latency."""
        # Connect to the mock server
        client = await AsyncioADSClient.connected_to(
            target_ip=perf_test_server.host,
            target_ams_net_id=perf_test_server.local_netid.to_string(),
            target_ams_port=IO_SERVER_PORT,
            ads_port=perf_test_server.port,
        )

        # Define the test sample population
        num_operations = 100

        # Create a sequence of random symbol/value pairs for testing
        test_symbols: list[tuple[AdsSymbol, Any]] = []
        for _ in range(num_operations):
            test_symbols.append(perf_test_server.get_random_symbol_value_pair())
        assert len(test_symbols) == num_operations

        try:
            # First warmup call to confirm functionality
            s0, v0 = test_symbols[0]
            await client.write_ads_symbol(s0, v0)

            # Latency measurements
            latencies = []
            for symbol, value in test_symbols:
                start = time.perf_counter()
                await client.write_ads_symbol(symbol, value)
                latency = time.perf_counter() - start
                latencies.append(latency)

            # Calculate statistics
            avg_latency = sum(latencies) / len(latencies)
            min_latency = min(latencies)
            max_latency = max(latencies)

            # Verify reasonable latencies (should be sub-second for mock)
            assert avg_latency < 1.0
            assert min_latency < max_latency

            # Report results to stdout
            print(
                f"Write symbol -- latency test results over {num_operations} samples:\n"
                + f"Avg: {avg_latency * 1000:.2f}ms,\n"
                + f"Min: {min_latency * 1000:.2f}ms,\n"
                + f"Max: {max_latency * 1000:.2f}ms"
            )

        finally:
            # Disconnect the client
            await client.close()

    @pytest.mark.asyncio
    async def test_read_write_latency(self, perf_test_server: MockADSServer):
        """Measure symbol combined read/write operation latency."""
        # Connect to the mock server
        client = await AsyncioADSClient.connected_to(
            target_ip=perf_test_server.host,
            target_ams_net_id=perf_test_server.local_netid.to_string(),
            target_ams_port=IO_SERVER_PORT,
            ads_port=perf_test_server.port,
        )

        # Define the test sample population
        num_operations = 100

        # Create a sequence of random symbol/value pairs for testing
        test_symbols: list[tuple[AdsSymbol, Any]] = []
        for _ in range(num_operations):
            test_symbols.append(perf_test_server.get_random_symbol_value_pair())
        assert len(test_symbols) == num_operations

        try:
            # First warmup call to confirm functionality
            s0, v0 = test_symbols[0]
            await client.readwrite_ads_symbol(s0, v0)

            # Latency measurements
            latencies = []
            for symbol, value in test_symbols:
                start = time.perf_counter()
                await client.readwrite_ads_symbol(symbol, value)
                latency = time.perf_counter() - start
                latencies.append(latency)

            # Calculate statistics
            avg_latency = sum(latencies) / len(latencies)
            min_latency = min(latencies)
            max_latency = max(latencies)

            # Verify reasonable latencies (should be sub-second for mock)
            assert avg_latency < 1.0
            assert min_latency < max_latency

            # Report results to stdout
            print(
                "Read/Write symbol -- "
                + f"latency test results over {num_operations} samples:\n"
                + f"Avg: {avg_latency * 1000:.2f}ms,\n"
                + f"Min: {min_latency * 1000:.2f}ms,\n"
                + f"Max: {max_latency * 1000:.2f}ms"
            )

        finally:
            # Disconnect the client
            await client.close()

    @pytest.mark.asyncio
    async def test_notification_subscription_latency(
        self, perf_test_server: MockADSServer
    ):
        """Measure symbol notification subscription latency."""
        # Connect to the mock server
        client = await AsyncioADSClient.connected_to(
            target_ip=perf_test_server.host,
            target_ams_net_id=perf_test_server.local_netid.to_string(),
            target_ams_port=IO_SERVER_PORT,
            ads_port=perf_test_server.port,
        )

        # Define the test sample population
        num_operations = 100

        # Create a sequence of discrete random symbols for testing notifications
        assert num_operations <= len(perf_test_server.symbols), (
            "Not enough symbol definitions in the mock server "
            + "for the requested number of subscriptions; "
            + f"current limit is {len(perf_test_server.symbols)}"
        )
        test_symbols: list[AdsSymbol] = []
        while len(test_symbols) < num_operations:
            test_symbols.append(perf_test_server.get_random_symbol())
            # Remove duplicates from the current test sample list
            test_symbols = list(set(test_symbols))
        assert len(test_symbols) == num_operations

        try:
            # Latency measurements
            latencies = []
            handles = []
            for symbol in test_symbols:
                start = time.perf_counter()
                handle = await client.add_device_notification(
                    symbol=symbol,
                    max_delay_ms=1000,
                    cycle_time_ms=100,
                )
                latency = time.perf_counter() - start
                latencies.append(latency)
                handles.append(handle)

            # Cleanup notification subscriptions
            for handle in handles:
                await client.delete_device_notification(handle)

            # Calculate statistics
            avg_latency = sum(latencies) / len(latencies)
            min_latency = min(latencies)
            max_latency = max(latencies)

            # Verify reasonable latencies (should be sub-second for mock)
            assert avg_latency < 1.0
            assert min_latency < max_latency

            # Report results to stdout
            print(
                "Symbol notification subscription -- "
                + f"latency test results over {num_operations} samples:\n"
                + f"Avg: {avg_latency * 1000:.2f}ms,\n"
                + f"Min: {min_latency * 1000:.2f}ms,\n"
                + f"Max: {max_latency * 1000:.2f}ms"
            )

        finally:
            # Disconnect the client
            await client.close()


# ===================================================================
# Throughput Tests
# ===================================================================


class TestOperationThroughput:
    """Test and measure operation throughput."""

    @pytest.mark.asyncio
    async def test_sequential_read_throughput(self, perf_test_server: MockADSServer):
        """Measure sequential read operations throughput."""
        # Connect to the mock server
        client = await AsyncioADSClient.connected_to(
            target_ip=perf_test_server.host,
            target_ams_net_id=perf_test_server.local_netid.to_string(),
            target_ams_port=IO_SERVER_PORT,
            ads_port=perf_test_server.port,
        )

        # Define the test sample population
        num_operations = 1000

        # Create a sequence of random symbols for testing
        test_symbols: list[AdsSymbol] = []
        for _ in range(num_operations):
            test_symbols.append(perf_test_server.get_random_symbol())
        assert len(test_symbols) == num_operations

        try:
            # Throughput measurements for single-threaded read operations
            start = time.perf_counter()
            for symbol in test_symbols:
                await client.read_ads_symbol(symbol)
            elapsed = time.perf_counter() - start

            # Calculate statistics and verify reasonable results
            throughput = num_operations / elapsed
            assert throughput > 0

            # Report results to stdout
            print(
                "Throughput test results for sequential read:\n"
                + f"{throughput:.1f} ops/sec based on {num_operations} samples"
            )
        finally:
            # Disconnect the client
            await client.close()

    @pytest.mark.asyncio
    async def test_sequential_write_throughput(self, perf_test_server: MockADSServer):
        """Measure sequential write operations throughput."""
        # Connect to the mock server
        client = await AsyncioADSClient.connected_to(
            target_ip=perf_test_server.host,
            target_ams_net_id=perf_test_server.local_netid.to_string(),
            target_ams_port=IO_SERVER_PORT,
            ads_port=perf_test_server.port,
        )

        # Define the test sample population
        num_operations = 1000

        # Create a sequence of random symbol/value pairs for testing
        test_symbols: list[tuple[AdsSymbol, Any]] = []
        for _ in range(num_operations):
            test_symbols.append(perf_test_server.get_random_symbol_value_pair())
        assert len(test_symbols) == num_operations

        try:
            # Throughput measurements for single-threaded write operations
            start = time.perf_counter()
            for symbol, value in test_symbols:
                await client.write_ads_symbol(symbol, value)
            elapsed = time.perf_counter() - start

            # Calculate statistics and verify reasonable results
            throughput = num_operations / elapsed
            assert throughput > 0

            # Report results to stdout
            print(
                "Throughput test results for sequential write:\n"
                + f"{throughput:.1f} ops/sec based on {num_operations} samples"
            )

        finally:
            # Disconnect the client
            await client.close()

    @pytest.mark.asyncio
    async def test_combined_read_write_operation_throughput(
        self, perf_test_server: MockADSServer
    ):
        """Measure combined read/write operation throughput."""
        # Connect to the mock server
        client = await AsyncioADSClient.connected_to(
            target_ip=perf_test_server.host,
            target_ams_net_id=perf_test_server.local_netid.to_string(),
            target_ams_port=IO_SERVER_PORT,
            ads_port=perf_test_server.port,
        )

        # Define the test sample population
        num_operations = 500

        # Create a sequence of random symbol/value pairs for testing
        test_symbols: list[tuple[AdsSymbol, Any]] = []
        for _ in range(num_operations):
            test_symbols.append(perf_test_server.get_random_symbol_value_pair())
        assert len(test_symbols) == num_operations

        try:
            # Throughput measurements for single-threaded read/write operations
            start = time.perf_counter()
            for symbol, value in test_symbols:
                await client.readwrite_ads_symbol(symbol, value)
            elapsed = time.perf_counter() - start

            # Calculate statistics and verify reasonable results
            throughput = num_operations / elapsed
            assert throughput > 0

            # Report results to stdout
            print(
                "Throughput test results for sequential combined read/write:\n"
                + f"{throughput:.1f} ops/sec based on {num_operations} samples"
            )

        finally:
            # Disconnect the client
            await client.close()

    @pytest.mark.asyncio
    async def test_concurrent_read_throughput(self, perf_test_server: MockADSServer):
        """Measure concurrent read operations throughput."""
        # Connect to the mock server
        client = await AsyncioADSClient.connected_to(
            target_ip=perf_test_server.host,
            target_ams_net_id=perf_test_server.local_netid.to_string(),
            target_ams_port=IO_SERVER_PORT,
            ads_port=perf_test_server.port,
        )

        # Define the test sample population
        num_tasks = 100
        num_operations = 10

        # Create a sequence of random symbols for testing
        test_symbols: list[AdsSymbol] = []
        for _ in range(num_operations):
            test_symbols.append(perf_test_server.get_random_symbol())
        assert len(test_symbols) == num_operations

        try:
            # Read task definition
            async def read_task():
                for symbol in test_symbols:
                    await client.read_ads_symbol(symbol)

            # Throughput measurements for parallel read operations
            start = time.perf_counter()
            await asyncio.gather(*[read_task() for _ in range(num_tasks)])
            elapsed = time.perf_counter() - start

            # Calculate statistics and verify reasonable results
            total_operations = num_tasks * num_operations
            throughput = total_operations / elapsed
            assert throughput > 0

            # Report results to stdout
            print(
                "Throughput test results for concurrent read operations:\n"
                + f"{throughput:.1f} ops/sec based on {num_tasks} tasks "
                + f"with {num_operations} samples each"
            )
        finally:
            # Disconnect the client
            await client.close()

    @pytest.mark.asyncio
    async def test_mixed_operations_throughput(self, perf_test_server: MockADSServer):
        """Measure throughput of mixed read and write operations."""
        # Connect to the mock server
        client = await AsyncioADSClient.connected_to(
            target_ip=perf_test_server.host,
            target_ams_net_id=perf_test_server.local_netid.to_string(),
            target_ams_port=IO_SERVER_PORT,
            ads_port=perf_test_server.port,
        )

        # Define the test sample population
        num_operations = 500

        # Create a sequence of random symbol/value pairs for testing
        test_symbols: list[tuple[AdsSymbol, Any]] = []
        for _ in range(num_operations):
            test_symbols.append(perf_test_server.get_random_symbol_value_pair())
        assert len(test_symbols) == num_operations

        try:
            # Throughput measurements for mixed type operations
            start = time.perf_counter()
            for i, (symbol, value) in enumerate(test_symbols):
                if i % 2 == 0:
                    await client.read_ads_symbol(symbol)
                else:
                    await client.write_ads_symbol(symbol, value)
            elapsed = time.perf_counter() - start

            # Calculate statistics and verify reasonable results
            throughput = num_operations / elapsed
            assert throughput > 0

            # Report results to stdout
            print(
                "Throughput test results for mixed symbol operations:\n"
                + f"{throughput:.1f} ops/sec based on {num_operations} samples"
            )

        finally:
            # Disconnect the client
            await client.close()


# ===================================================================
# Scalability Tests
# ===================================================================


class TestScalability:
    """Test scalability with increasing load."""

    # # TO DO: if multiple clients can be mocked, test throughput
    # # with concurrent connections
    # @pytest.mark.asyncio
    # async def test_scalability_concurrent_clients(
    #     self, perf_test_server: MockADSServer
    # ):
    #     """Test scalability with increasing number of concurrent clients."""
    #     scalability_results = {}

    #     # Define the test sample populations
    #     num_clients = [1, 5, 10, 20]
    #     num_operations = 100

    #     # Scalability measurements for different number of clients
    #     for nclients in num_clients:
    #         clients = []
    #         try:
    #             # Create clients
    #             for i in range(nclients):
    #                 # connection to server
    #                 clients.append(client)

    #             # Define client operation task to measure throughput
    #             async def client_operations(client):
    #                 symbol = perf_test_server.get_random_symbol()
    #                 for _ in range(num_operations):
    #                     await client.read_ads_symbol(symbol)

    #             # Scalability measurements with multiple clients
    #             start = time.perf_counter()
    #             await asyncio.gather(*[client_operations(c) for c in clients])
    #             elapsed = time.perf_counter() - start

    #             # Calculate statistics for each client load
    #             throughput = (nclients * num_operations) / elapsed
    #             scalability_results[nclients] = throughput

    #         finally:
    #             # Disconnect the clients
    #             for client in clients:
    #                 await client.close()

    #     # Verify reasonable results are achieved
    #     assert all(tp > 0 for tp in scalability_results.values())

    #     # Report results to stdout
    #     print("Scalability test results for different client loads:\n")
    #     for num_clients, throughput in scalability_results.items():
    #         print(
    #             f"Symbol read: {throughput:.1f} ops/sec "
    #             + f"when {num_clients} clients are concurrently connected "
    #             + f"to the server"
    #         )

    @pytest.mark.asyncio
    async def test_scalability_concurrent_notifications(
        self, perf_test_server: MockADSServer
    ):
        """Test scalability with increasing active notifications."""
        # Connect to the mock server
        client = await AsyncioADSClient.connected_to(
            target_ip=perf_test_server.host,
            target_ams_net_id=perf_test_server.local_netid.to_string(),
            target_ams_port=IO_SERVER_PORT,
            ads_port=perf_test_server.port,
        )

        # Define the test sample populations
        num_notifications = [10, 50, 100]
        num_operations = 100

        # Create a sequence of discrete random symbols for testing notifications
        assert max(num_notifications) <= len(perf_test_server.symbols), (
            "Not enough symbol definitions in the mock server "
            + "for the requested number of subscriptions; "
            + f"current limit is {len(perf_test_server.symbols)}"
        )
        test_symbols: list[AdsSymbol] = []
        while len(test_symbols) < max(num_notifications):
            test_symbols.append(perf_test_server.get_random_symbol())
            # Remove duplicates from the current test sample list
            test_symbols = list(set(test_symbols))
        assert len(test_symbols) == max(num_notifications)

        # Scalability measurements for different notification loads
        try:
            scalability_results = {}
            for n in num_notifications:
                handles = []
                symbols = test_symbols[:n]

                # Create notification subscriptions
                for symbol in symbols:
                    handle = await client.add_device_notification(
                        symbol=symbol,
                        max_delay_ms=1000,
                        cycle_time_ms=100,
                    )
                    handles.append(handle)

                # Measure read throughput with notifications active
                s = perf_test_server.get_random_symbol()
                start = time.perf_counter()
                for _ in range(num_operations):
                    await client.read_ads_symbol(s)
                elapsed = time.perf_counter() - start

                # Calculate statistics for each load
                throughput = num_operations / elapsed
                scalability_results[num_notifications] = throughput

                # Cleanup notification subscriptions
                for handle in handles:
                    await client.delete_device_notification(handle)

            # Verify reasonable results are achieved
            assert all(tp > 0 for tp in scalability_results.values())

            # Report results to stdout
            print("Scalability test results for different notification loads:\n")
            for num_notifs, throughput in scalability_results.items():
                print(
                    f"Symbol read: {throughput:.1f} ops/sec "
                    + f"when concurrently subscribed to {num_notifs} notifications"
                )

        finally:
            # Disconnect the client
            await client.close()


# ===================================================================
# Memory and Resource Tests
# ===================================================================


class TestMemoryAndResources:
    """Test memory usage and resource management."""

    @pytest.mark.asyncio
    async def test_symbol_creation_efficiency(self):
        """Test efficiency of symbol creation."""
        # Define the test sample population
        num_symbols = 1000

        # Performance measurement through symbol instantiation
        start_time = time.perf_counter()
        symbols = []
        for i in range(num_symbols):
            symbol = AdsSymbol(
                parent_id=i % 10,
                name=f"Symbol{i}",
                dtype=np.int32,
                size=random.randint(1, 100),
                group=0x3000,
                offset=i * 4,
                comment=f"Symbol {i} comment",
            )
            symbols.append(symbol)
        elapsed = time.perf_counter() - start_time

        # Calculate statistics
        throughput = num_symbols / elapsed

        # Verify reasonable results are achieved
        assert throughput > 0
        assert len(symbols) == num_symbols

        # Report results to stdout
        print(
            "Throughput test results for symbol creation:\n"
            + f"{throughput:.1f} symbols/sec based on {num_symbols} samples"
        )

    @pytest.mark.asyncio
    async def test_data_accumulation(self, perf_test_server: MockADSServer):
        """Test data accumulation without memory leaks."""
        # Connect to the mock server
        client = await AsyncioADSClient.connected_to(
            target_ip=perf_test_server.host,
            target_ams_net_id=perf_test_server.local_netid.to_string(),
            target_ams_port=IO_SERVER_PORT,
            ads_port=perf_test_server.port,
        )

        # Define the test sample population
        num_symbols = 1000

        # Create a sequence of random symbols for testing
        test_symbols: list[AdsSymbol] = []
        for _ in range(num_symbols):
            test_symbols.append(perf_test_server.get_random_symbol())
        assert len(test_symbols) == num_symbols

        # Memory leak measurement through data accumulation over time
        try:
            # Perform successive symbol reads and cache the data
            data_buffer = []
            for symbol in test_symbols:
                result = await client.read_ads_symbol(symbol)
                data_buffer.append(result)

            # If no memory leak, data should have been accumulated
            assert len(data_buffer) == num_symbols

        finally:
            # Disconnect the client
            await client.close()


# ===================================================================
# Stress Tests
# ===================================================================


class TestStressConditions:
    """Test behavior under stress conditions."""

    @pytest.mark.asyncio
    async def test_rapid_connection_cycles(self, perf_test_server: MockADSServer):
        """Test overhead from rapid connect/disconnect cycles."""
        # Define the test sample population
        num_cycles = 50

        # Stress measurement through the repeated cycling of client connections
        start = time.perf_counter()
        for _ in range(num_cycles):
            # Connect to the mock server
            client = await AsyncioADSClient.connected_to(
                target_ip=perf_test_server.host,
                target_ams_net_id=perf_test_server.local_netid.to_string(),
                target_ams_port=IO_SERVER_PORT,
                ads_port=perf_test_server.port,
            )
            # Quick ads operation
            await client._get_ioserver_states()
            # Disconnect the client
            await client.close()
        elapsed = time.perf_counter() - start

        # Calculate statistics
        cycle_time = elapsed / num_cycles

        # Report results to stdout
        print(
            "Stress test result for rapid client connection cycling:\n"
            + f"{cycle_time * 1000:.2f} ms/cycle based on {num_cycles} samples"
        )

    @pytest.mark.asyncio
    async def test_high_concurrency(self, perf_test_server: MockADSServer):
        """Test behaviour under high concurrency."""
        # Connect to the mock server
        client = await AsyncioADSClient.connected_to(
            target_ip=perf_test_server.host,
            target_ams_net_id=perf_test_server.local_netid.to_string(),
            target_ams_port=IO_SERVER_PORT,
            ads_port=perf_test_server.port,
        )

        # Define the test sample populations
        num_tasks = 200
        num_operations = 10

        # Stress measurement through handling of concurrent tasks with high read load
        try:
            # Create many concurrent tasks
            async def task():
                # Create a sequence of random symbols for testing
                test_symbols: list[AdsSymbol] = []
                for _ in range(num_operations):
                    test_symbols.append(perf_test_server.get_random_symbol())
                    assert len(test_symbols) == num_operations
                # Setup a read operation on each symbol
                for symbol in test_symbols:
                    await client.read_ads_symbol(symbol)

            # Measure total throughput of the tasks
            start = time.perf_counter()
            await asyncio.gather(*[task() for _ in range(num_tasks)])
            elapsed = time.perf_counter() - start

            # Calculate statistics
            total_ops = num_tasks * num_operations
            throughput = total_ops / elapsed

            # Report results to stdout
            print(
                "Stress test result for high concurrency task requirement:\n"
                + f"{throughput:.1f} ops/sec based on {num_tasks} operation tasks "
                + f"with {num_operations} samples each"
            )

        finally:
            # Disconnect the client
            await client.close()

    @pytest.mark.asyncio
    async def test_sustained_load(self, perf_test_server: MockADSServer):
        """Test sustained load over extended period."""
        # Connect to the mock server
        client = await AsyncioADSClient.connected_to(
            target_ip=perf_test_server.host,
            target_ams_net_id=perf_test_server.local_netid.to_string(),
            target_ams_port=IO_SERVER_PORT,
            ads_port=perf_test_server.port,
        )

        # Define the test stress constraint (i.e. sustained load)
        duration = 2.0

        # Stress measurement through the repeated read operation over an extended period
        try:
            total_operations = 0
            symbol = perf_test_server.get_random_symbol()
            start = time.perf_counter()
            while time.perf_counter() - start < duration:
                await client.read_ads_symbol(symbol)
                total_operations += 1
            elapsed = time.perf_counter() - start

            # Calculate statistics
            throughput = total_operations / elapsed

            # Report results to stdout
            print(
                "Stress test result for sustained read operation over a long period:\n"
                + f"{throughput:.1f} ops/sec based on {total_operations} total "
                + f"operations over a {duration}s extended period"
            )

        finally:
            # Disconnect the client
            await client.close()


# ===================================================================
# Comparative Performance Tests
# ===================================================================


class TestComparativePerformance:
    """Test comparative performance between different operations."""

    @pytest.mark.asyncio
    async def test_read_vs_write_performance(self, perf_test_server: MockADSServer):
        """Compare read vs write speed performance."""
        # Connect to the mock server
        client = await AsyncioADSClient.connected_to(
            target_ip=perf_test_server.host,
            target_ams_net_id=perf_test_server.local_netid.to_string(),
            target_ams_port=IO_SERVER_PORT,
            ads_port=perf_test_server.port,
        )

        # Define the test sample population
        num_operations = 500

        # Create a sequence of random symbol/value pairs for testing
        test_symbols: list[tuple[AdsSymbol, Any]] = []
        for _ in range(num_operations):
            test_symbols.append(perf_test_server.get_random_symbol_value_pair())
        assert len(test_symbols) == num_operations

        # Comparative performance measurement between symbol various operations
        try:
            # Throughput measurement with successive read operations
            start = time.perf_counter()
            for symbol, _ in test_symbols:
                await client.read_ads_symbol(symbol)
            read_elapsed = time.perf_counter() - start
            read_throughput = num_operations / read_elapsed

            # Throughput measurement with successive write operations
            start = time.perf_counter()
            for symbol, value in test_symbols:
                await client.write_ads_symbol(symbol, value)
            write_elapsed = time.perf_counter() - start
            write_throughput = num_operations / write_elapsed
            write_ratio = read_throughput / write_throughput

            # Throughput measurement with successive read/write operations
            start = time.perf_counter()
            for symbol, value in test_symbols:
                await client.readwrite_ads_symbol(symbol, value)
            readwrite_elapsed = time.perf_counter() - start
            readwrite_throughput = num_operations / readwrite_elapsed
            readwrite_ratio = read_throughput / readwrite_throughput

            # Report results to stdout
            print(
                "Comparative performance test results for symbol operations "
                + f"based on {num_operations} samples:\n"
                + f"Symbol read: {read_throughput:.1f} ops/sec\n"
                + f"Symbol write: {write_throughput:.1f} ops/sec\n"
                + f"Symbol read/write: {readwrite_throughput:.1f} ops/sec\n"
                + f"Read operation is {write_ratio:.1f}x the write performance\n"
                + f"Read operation is {readwrite_ratio:.1f}x the read/write performance"
            )

        finally:
            # Disconnect the client
            await client.close()

    @pytest.mark.asyncio
    async def test_sequential_vs_concurrent_performance(
        self, perf_test_server: MockADSServer
    ):
        """Compare sequential vs concurrent speed performance."""
        # Connect to the mock server
        client = await AsyncioADSClient.connected_to(
            target_ip=perf_test_server.host,
            target_ams_net_id=perf_test_server.local_netid.to_string(),
            target_ams_port=IO_SERVER_PORT,
            ads_port=perf_test_server.port,
        )

        # Define the test sample populations
        num_operations = 100
        num_tasks = 10
        assert num_operations % num_tasks == 0, "Test populations must be proportional"
        task_smpl_slice = num_operations // num_tasks

        # Create a sequence of random symbols for testing
        test_symbols: list[AdsSymbol] = []
        for _ in range(num_operations):
            test_symbols.append(perf_test_server.get_random_symbol())
            assert len(test_symbols) == num_operations

        # Comparative performance measurement bewtween read operation methods
        try:
            # Throughput measurement with sequential read operations
            start = time.perf_counter()
            for symbol in test_symbols:
                await client.read_ads_symbol(symbol)
            seq_elapsed = time.perf_counter() - start
            seq_throughput = num_operations / seq_elapsed

            # Throughput measurement with concurrent read operations
            async def read_task(index: int):
                start = index * task_smpl_slice
                end = (index + 1) * task_smpl_slice
                for symbol in test_symbols[start:end]:
                    await client.read_ads_symbol(symbol)

            start = time.perf_counter()
            await asyncio.gather(*[read_task(i) for i in range(num_tasks)])
            conc_elapsed = time.perf_counter() - start
            conc_throughput = num_operations / conc_elapsed

            # Calculate statistics
            ratio = conc_throughput / seq_throughput

            # Report results to stdout
            print(
                "Comparative performance test results for different query modes "
                + f"based on {num_operations} read samples:\n"
                + f"Sequential read throughput: {seq_throughput:.1f} ops/sec\n"
                + f"Concurrent read throughput ({num_tasks} threads): "
                + f"{conc_throughput:.1f} ops/sec\n"
                + f"Concurrent mode is {ratio:.1f}x the sequential performance"
            )

        finally:
            # Disconnect the client
            await client.close()


# ===================================================================
# Bottleneck Identification Tests
# ===================================================================


class TestBottleneckIdentification:
    """Test to identify performance bottlenecks."""

    @pytest.mark.asyncio
    async def test_payload_size_impact(self, perf_test_server: MockADSServer):
        """Test impact of payload size on throughput performance."""
        # Connect to the mock server
        client = await AsyncioADSClient.connected_to(
            target_ip=perf_test_server.host,
            target_ams_net_id=perf_test_server.local_netid.to_string(),
            target_ams_port=IO_SERVER_PORT,
            ads_port=perf_test_server.port,
        )

        # Define the test sample populations
        payload_sizes = [1, 4, 16, 64, 256, 1024]
        num_operations = 100

        # Create ads test symbols with different data sizes on the mock server
        for size in payload_sizes:
            perf_test_server.set_symbol_data(
                name=f"BottleneckSymbol{size}", value=np.arange(size, dtype=np.uint8)
            )

        # Performance measurement for symbol with different data size
        try:
            results = {}
            for size in payload_sizes:
                # Select an ads symbol with the correct data size
                symbol = perf_test_server.get_symbol_by_name(f"BottleneckSymbol{size}")
                assert symbol is not None, (
                    f"No test symbol suitable for payload size {size} on mock server"
                )

                # Throughput measurement with multiple read operations
                start = time.perf_counter()
                for _ in range(num_operations):
                    await client.read_ads_symbol(symbol)
                elapsed = time.perf_counter() - start

                # Calculate statistics and record
                throughput = num_operations / elapsed
                results[size] = throughput

            # Report results to stdout
            print(
                "Performance test results for different payload sizes "
                + f"based on {num_operations} read samples:\n"
            )
            for size, throughput in results.items():
                print(f"{throughput:.1f} ops/sec throughput for {size} bytes payload\n")

        finally:
            # Disconnect the client
            await client.close()

    @pytest.mark.asyncio
    async def test_offset_distribution_impact(self, perf_test_server: MockADSServer):
        """Test impact of offset distribution on performance related to memory access \
            pattern effects"""
        # Connect to the mock server
        client = await AsyncioADSClient.connected_to(
            target_ip=perf_test_server.host,
            target_ams_net_id=perf_test_server.local_netid.to_string(),
            target_ams_port=IO_SERVER_PORT,
            ads_port=perf_test_server.port,
        )

        # Define the test sample population
        num_operations = 100

        # Create ads test symbols with different data sizes on the mock server
        for i in range(num_operations):
            perf_test_server.set_symbol_data(
                name=f"OffsetSymbol{i}", value=np.int32(i), offset=(i * 256)
            )

        # Performance measurement for symbol with different offset index
        try:
            # Throughput measurement with same symbol, i.e. same offset
            symbol = perf_test_server.get_random_symbol()
            start = time.perf_counter()
            for _ in range(num_operations):
                await client.read_ads_symbol(symbol)
            same_elapsed = time.perf_counter() - start
            same_throughput = num_operations / same_elapsed

            # Throughput measurement for symbols with varying offset
            test_symbols: list[AdsSymbol] = []
            for i in range(num_operations):
                name = f"OffsetSymbol{i}"
                symbol = perf_test_server.get_symbol_by_name(name)
                assert symbol is not None, (
                    f"No test symbol with name {name} is available on mock server"
                )
                test_symbols.append(symbol)
            start = time.perf_counter()
            for symbol in test_symbols:
                await client.read_ads_symbol(symbol)
            varying_elapsed = time.perf_counter() - start
            varying_throughput = num_operations / varying_elapsed

            # Report results to stdout
            print(
                "Performance test results for different symbol offset distribution "
                + f"based on {num_operations} read samples:\n"
                + f"{same_throughput:.1f} ops/sec throughput for same offset symbol\n"
                + f"{varying_throughput:.1f} ops/sec throughput for varying offset "
                + "symbols"
            )

        finally:
            await client.close()



"""
Provides two pytest fixtures for the mock ADS server infrastructure:
- `mock_ads_server` - Async fixture that starts and stops the server automatically
- `mock_server_address` - Fixture that returns the server address for tests
"""

import os
from collections.abc import AsyncGenerator
from typing import Any

import pytest

from mock_server import MockADSServer

# Prevent pytest from catching exceptions when debugging in vscode so that break on
# exception works correctly (see: https://github.com/pytest-dev/pytest/issues/7409)
if os.getenv("PYTEST_RAISE", "0") == "1":

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call: pytest.CallInfo[Any]):
        if call.excinfo is not None:
            raise call.excinfo.value
        else:
            raise RuntimeError(
                f"{call} has no exception data, an unknown error has occurred"
            )

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo: pytest.ExceptionInfo[Any]):
        raise excinfo.value


@pytest.fixture
async def mock_ads_server() -> AsyncGenerator[MockADSServer]:
    """Fixture that provides a running mock ADS server for testing."""
    server = MockADSServer(host="127.0.0.1", port=48898)
    await server.start()
    yield server
    await server.stop()


@pytest.fixture
def mock_server_address() -> tuple[str, int]:
    """Fixture that provides the mock server address."""
    return ("127.0.0.1", 48898)



"""
This file checks that all the example boilerplate text has been removed.
It can be deleted when all the contained tests pass
"""

from importlib.metadata import metadata  # noqa
from pathlib import Path

ROOT = Path(__file__).parent.parent


def skeleton_check(check: bool, text: str):
    if ROOT.name == "python3-pip-skeleton" or str(ROOT) == "/project":
        # In the skeleton module the check should fail
        check = not check
        text = f"Skeleton didn't raise: {text}"
    if check:
        raise AssertionError(text)


def assert_not_contains_text(path: str, text: str, explanation: str):
    full_path = ROOT / path
    if full_path.exists():
        contents = full_path.read_text().replace("\n", " ")
        skeleton_check(text in contents, f"Please change ./{path} {explanation}")


# pyproject.toml
def test_module_summary():
    summary = metadata("fastcs_catio")["summary"]
    skeleton_check(
        "One line description of your module" in summary,
        "Please change project.description in ./pyproject.toml "
        "to be a one line description of your module",
    )


# README
def test_changed_readme_intro():
    assert_not_contains_text(
        "README.rst",
        "This is where you should write a short paragraph",
        "to include an intro on what your module does",
    )


def test_removed_adopt_skeleton():
    assert_not_contains_text(
        "README.rst",
        "This project contains template code only",
        "remove the note at the start",
    )


def test_changed_readme_body():
    assert_not_contains_text(
        "README.rst",
        "This is where you should put some images or code snippets",
        "to include some features and why people should use it",
    )



from enum import EJECT, Enum, Flag

import numpy as np

# Python Standard Encodings
# https://docs.python.org/3.8/library/codecs.html#standard-encodings

# A STRING constant is a string enclosed by single quotation marks.
# The characters are encoded according to the Windows 1252 character set.
# https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_plc_intro/2529327243.html&id=
TWINCAT_STRING_ENCODING = "cp1252"


class CommandId(np.uint16, Enum):
    """
    Command ID
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/115847307.html&id=7738940192708835096
    """

    ADSSRVID_READDEVICEINFO = 0x1
    """Reads the name and the version number of the ADS device."""
    ADSSRVID_READ = 0x2
    """With ADS Read data can be read from an ADS device."""
    ADSSRVID_WRITE = 0x3
    """With ADS Write data can be written to an ADS device."""
    ADSSRVID_READSTATE = 0x4
    """Reads the ADS status and the device status of an ADS device."""
    ADSSRVID_WRITECTRL = 0x5
    """Changes the ADS status and the device status of an ADS device."""
    ADSSRVID_ADDDEVICENOTE = 0x6
    """A notification is created in an ADS device."""
    ADSSRVID_DELETEDEVICENOTE = 0x7
    """One before defined notification is deleted in an ADS device."""
    ADSSRVID_DEVICENOTE = 0x8
    """Data will carry forward independently from an ADS device to a Client."""
    ADSSRVID_READWRITE = 0x9
    """With ADS ReadWrite data will be written to an ADS device.
    Additionally, data can be read from the ADS device."""


class SystemServiceCommandId(np.uint32, Enum):
    """
    Specification of the ADS system services
    """

    ADSSVCID_READSERVICEINFO = 0x1
    ADSSVCID_ADDROUTE = 0x6
    ADSSVCID_DELROUTE = 0xB001
    ADSSCVID_RESPONSE = 0x80000000


class UDPTag(np.uint16, Enum):
    """
    UDP Tags for ADS Router configuration
    """

    PASSWORD = 2
    HOSTNAME = 5
    HOSTNETID = 7
    ROUTENAME = 12
    USERNAME = 13


class StateFlag(np.uint16, Enum):
    """
    State Flags
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/115847307.html&id=7738940192708835096
    """

    AMSCMDSF_RESPONSE = 0x1
    AMSCMDSF_ADSCMD = 0x4


class AdsState(np.uint16, Enum):
    """
    Status of the ADS interface of the ADS device
    https://infosys.beckhoff.com/english.php?content=../content/1033/tcplclib_tc2_system/31084171.html&id=
    """

    ADSSTATE_INVALID = 0
    """Invalid state"""
    ADSSTATE_IDLE = 1
    """Idle state"""
    ADSSTATE_RESET = 2
    """Reset state"""
    ADSSTATE_INIT = 3
    """Initialised state"""
    ADSSTATE_START = 4
    """Started"""
    ADSSTATE_RUN = 5
    """Running"""
    ADSSTATE_STOP = 6
    """Stopped"""
    ADSSTATE_SAVECFG = 7
    """Save configuration"""
    ADSSTATE_LOADCFG = 8
    """Load configuration"""
    ADSSTATE_POWERFAILURE = 9
    """Power failure"""
    ADSSTATE_POWERGOOD = 10
    """Power good"""
    ADSSTATE_ERROR = 11
    """Error state"""
    ADSSTATE_SHUTDOWN = 12
    """Shutting down"""
    ADSSTATE_SUSPEND = 13
    """Suspended"""
    ADSSTATE_RESUME = 14
    """Resumed"""
    ADSSTATE_CONFIG = 15
    """System is in config mode"""
    ADSSTATE_RECONFIG = 16
    """System should restart in config mode"""
    ADSSTATE_STOPPING = 17
    """Stopping state"""
    ADSSTATE_INCOMPATIBLE = 18
    """Incompatible state"""
    ADSSTATE_EXCEPTION = 19
    """Exception state"""
    ADSSTATE_MAXSTATES = 20
    """Maximum number of available ads states"""


class DeviceStateMachine(np.uint16, Enum):
    """
    EtherCAT device state machine
    https://infosys.beckhoff.com/english.php?content=../content/1033/tcsystemmanager/1089026187.html&id=
    """

    DEVSTATE_INIT = 0x00
    """Init state"""
    DEVSTATE_PREOP = 0x02
    """Pre-operational state"""
    DEVSTATE_BOOTSTRAP = 0x03
    """Bootstrap state"""
    DEVSTATE_SAFEOP = 0x04
    """Safe-operational state"""
    DEVSTATE_OP = 0x08
    """Operational state"""


class SlaveStateMachine(np.uint8, Enum):
    """
    EtherCAT slave state machine
    https://infosys.beckhoff.com/english.php?content=../content/1033/tcplclib_tc2_ethercat/57122443.html&id=
    """

    SLAVE_STATE_INIT = 0x00
    """Init state"""
    SLAVE_STATE_PREOP = 0x02
    """Pre-operational state"""
    SLAVE_STATE_BOOTSTRAP = 0x03
    """Bootstrap state"""
    SLAVE_STATE_SAFEOP = 0x04
    """Safe-operational state"""
    SLAVE_STATE_OP = 0x08
    """Operational state"""
    SLAVE_STATE_ERROR = 0x10
    """Error state"""
    SLAVE_STATE_INVALID_ID = 0x20
    """Invalid vendorId, productCode, revisionNb or serialNb"""
    SLAVE_STATE_INIT_ERROR = 0x40
    """Initialisation command error"""
    SLAVE_STATE_DISABLED = 0x80
    """Disabled slave"""


class SlaveLinkState(np.uint8, Enum):
    """
    EtherCAT slave link status
    https://infosys.beckhoff.com/english.php?content=../content/1033/tcplclib_tc2_ethercat/57122443.html&id=
    """

    SLAVE_LINK_STATE_OK = 0x00
    """Good ADS link state"""
    SLAVE_LINK_STATE_NOT_PRESENT = 0x01
    """No EtherCAT communication with the EtherCAT slave"""
    SLAVE_LINK_STATE_WITHOUT_COMM = 0x02
    """Error at port X; the port has a link, but no communication is possible via it."""
    SLAVE_LINK_STATE_MISSING_LINK = 0x04
    """Missing communication link at port X"""
    SLAVE_LINK_STATE_ADDITIONAL_LINK = 0x08
    """Additional communication link at port X"""
    SLAVE_LINK_STATE_PORT_A = 0x10
    """Port 0"""
    SLAVE_LINK_STATE_PORT_B = 0x20
    """Port 1"""
    SLAVE_LINK_STATE_PORT_C = 0x40
    """Port 2"""
    SLAVE_LINK_STATE_PORT_D = 0x80
    """Port 3"""


# TwinCAT 2 Reserved Index Groups
# https://infosys.beckhoff.com/english.php?content=../content/1033/tcplclibsystem/11827998603.html&id=
class IndexGroup(np.uint32, Flag, boundary=EJECT):  # type: ignore[misc]
    """
    Specification of the ADS system services
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/117463563.html&id=
    """

    ADSIGRP_MASTER_STATEMACHINE = 0x0003
    """Group used to access the state machine of the Master EtherCAT device"""
    ADSIGRP_MASTER_COUNT_SLAVE = 0x0006
    """Group used to access the number of configured slaves"""
    ADSIGRP_MASTER_SLAVE_ADDRESSES = 0x0007
    """Group used to access the fixed addresses of all slaves configured on a device"""
    ADSIGRP_SLAVE_STATEMACHINE = 0x0009
    """Group used to access the state machine of the slave terminal"""
    ADSIGRP_MASTER_FRAME_COUNTERS = 0x000C
    """Group used to access the frame counters of an EtherCAT device"""
    ADSIGRP_MASTER_SLAVE_IDENTITY = 0x0011
    """Group used to access the CANopen identity of an EtherCAT slave"""
    ADSIGRP_SLAVE_CRC_COUNTERS = 0x0012
    """Group used to access the CRC counters of all slaves configured on a device"""
    ADSIGR_IODEVICE_STATE_BASE = 0x5000
    """Group used to access information about the IO server"""
    ADSIGR_GET_SYMHANDLE_BYNAME = 0xF003
    """A handle (code word) is assigned to the name contained in the write data
    and is returned to the caller as a result ('symHdl')."""
    ADSIGR_GET_SYMVAL_BYHANDLE = 0xF005
    """Reads the value of the variable identified by 'symHdl'
    or assigns a value to the variable if not defined yet.
    (symHdl index offset range: 0x00000000 - 0xFFFFFFFF)"""
    ADSIGRP_RELEASE_SYMHANDLE = 0xF006
    """The code (handle) which is contained in the write data for an interrogated,
    named PLC variable is released."""
    ADSIGRP_IOIMAGE_RWIB = 0xF020
    """Group used to read and write input byte(s)
    (the associated index offset is byte offset, range: 0x0001F400 - 0xFFFFFFFF)"""
    ADSIGRP_IOIMAGE_RWIX = 0xF021
    """Group used to read and write input bit
    (the associated index offset contains the bit address calculated from:
    base offset (0xFA000) + (byte number * 8) + bit number
    range: 0x000FA000 - 0xFFFFFFFF)"""
    ADSIGRP_IOIMAGE_RISIZE = 0xF025
    """The size of the inputs in bytes"""
    ADSIGRP_IOIMAGE_RWOB = 0xF030
    """Group used to read and write output byte(s)
    (the associated index offset is byte offset, range: 0x0003E800 - 0xFFFFFFFF)"""
    ADSIGRP_IOIMAGE_RWOX = 0xF031
    """Group used to read and write output bit
    (the associated index offset contains the bit address calculated from:
    base offset (0x1F4000) + (byte number * 8) + bit number
    range: 0x001F4000 - 0xFFFFFFFF)"""
    ADSIGRP_IOIMAGE_RWOSIZE = 0xF035
    """The size of the outputs in bytes"""
    ADSIGRP_SUMUP_READ = 0xF080
    """Writes a list of multiple, separate AdsReadRequest sub-commands.
    Reads a list of return codes followed by the requested data."""
    ADSIGRP_SUMUP_WRITE = 0xF081
    """Writes a list of multiple, separate AdsWriteRequest sub-commands.
    Reads a list of return codes."""
    ADSIGRP_SUMUP_READWRITE = 0xF082
    """Writes a list of multiple, separate AdsReadWriteRequest sub-commands.
    Reads a list of return code and data length followed by the requested data."""
    ADSIGRP_SYM_UPLOAD = 0xF00B
    """Symbol uploads group, used to access the available symbol entries"""
    ADSIGRP_SYM_UPLOADINFO2 = 0xF00F
    """Symbol info uploads group, used to access info about the available symbols"""
    ADSIGRP_COE_LINK = 0xF302
    """IO link data channel, used to access the Can-over-EtherCAT (CoE) parameters"""


class CoEIndex(np.uint32, Enum):
    """
    Specification of the CAN-over-EtherCAT (CoE) parameter ranges
    https://infosys.beckhoff.com/english.php?content=../content/1033/ethercatsystem/2469073803.html&id=274367601078044781
    """

    ADS_COE_IDENTIFICATION_PARAMS = 0x1000
    """Range used to access the fixed identity information of a device/terminal"""
    ADS_COE_OPERATIONAL_PARAMS = 0x8000
    """Range used to access the operational and functional parameters"""


class DeviceType(np.uint16, Enum):
    """
    I/O device types
    https://infosys.beckhoff.com/english.php?content=../content/1033/tcplclib_tc2_iofunctions/59217419.html&id=
    """

    IODEVICETYPE_INVALID = 0
    """Device type not available (e.g. Invalid Interface Error response)"""
    IODEVICETYPE_ETHERCAT = 94
    """EtherCAT in direct mode"""


class AdsDataType(np.uint32, Enum):
    """
    Ads data type id
    https://infosys.beckhoff.com/english.php?content=../content/1033/tcplclib_tc2_utilities/35330059.html&id=
    """

    ADS_TYPE_VOID = 0
    """Reserved"""
    ADS_TYPE_INT16 = 2
    """Signed 16 bit integer (INT16)"""
    ADS_TYPE_INT32 = 3
    """Signed 32 bit integer (INT32)"""
    ADS_TYPE_REAL32 = 4
    """32 bit floating point number (REAL)"""
    ADS_TYPE_REAL64 = 5
    """64 bit floating point number (LREAL)"""
    ADS_TYPE_INT8 = 16
    """Signed 8 bit integer (INT8)"""
    ADS_TYPE_UINT8 = 17
    """Unsigned 8 bit integer (UINT8|BYTE)"""
    ADS_TYPE_UINT16 = 18
    """Unsigned 16 bit integer (UINT16|WORD)"""
    ADS_TYPE_UINT32 = 19
    """Unsigned 32 bit integer (UINT32|DWORD)"""
    ADS_TYPE_INT64 = 20
    """Signed 64 bit integer (INT64)"""
    ADS_TYPE_UINT64 = 21
    """Unsigned 64 bit integer (UINT64|LWORD)"""
    ADS_TYPE_STRING = 30
    """String type (STRING)"""
    ADS_TYPE_WSTRING = 31
    """Wide character type (WSTRING)"""
    ADS_TYPE_REAL80 = 32
    """Reserved"""
    ADS_TYPE_BIT = 33
    """Bit type (BIT)"""
    ADS_TYPE_MAXTYPES = 34
    """Maximum available type"""
    ADS_TYPE_BIGTYPE = 65
    """Structured type (STRUCT)"""


class SymbolFlag(np.uint32, Enum):
    """
    Ads symbol flag used as one of the symbol table entry property
    """

    ADS_SYMBOLFLAG_PERSISTENT = 1 << 0
    ADS_SYMBOLFLAG_BITVALUE = 1 << 1
    ADS_SYMBOLFLAG_REFERENCETO = 1 << 2
    ADS_SYMBOLFLAG_TYPEGUID = 1 << 3
    ADS_SYMBOLFLAG_TCCOMIFACEPTR = 1 << 4
    ADS_SYMBOLFLAG_READONLY = 1 << 5
    ADS_SYMBOLFLAG_CONTEXTMASK = 0xF00


class TransmissionMode(np.uint32, Enum):
    """
    C++ ADSTRANSMODE Enum
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_adsdll2/117558283.html&id=

    C++ AdsNotificationAttrib structure
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_adsdll2/117553803.html&id=
    """

    ADSTRANS_NOTRANS = 0
    """No notifications."""
    ADSTRANS_CLIENTCYCLE = 1
    """The notification's callback is invoked with the client cycle."""
    ADSTRANS_CLIENT1REQ = 2
    """The notification's callback function is invoked only once."""
    ADSTRANS_SERVERCYCLE = 3
    """The notification's callback function is invoked cyclically."""
    ADSTRANS_SERVERONCHA = 4
    """The notification's callback function is only invoked when the value changes."""


class ErrorCode(np.uint32, Enum):
    """
    ADS Return Codes
    https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro/374277003.html&id=4954945278371876402
    """

    ERR_NOERROR = 0x0
    """No error."""
    ERR_INTERNAL = 0x1
    """Internal error."""
    ERR_NORTIME = 0x2
    """No real time."""
    ERR_ALLOCLOCKEDMEM = 0x3
    """Allocation locked - memory error."""
    ERR_INSERTMAILBOX = 0x4
    """Mailbox full - the ADS message could not be sent.
    Reducing the number of ADS messages per cycle will help."""
    ERR_WRONGRECEIVEHMSG = 0x5
    """Wrong HMSG."""
    ERR_TARGETPORTNOTFOUND = 0x6
    """Target port not found - ADS server is not started or is not reachable."""
    ERR_TARGETMACHINENOTFOUND = 0x07
    """Target computer not found - AMS route was not found."""
    ERR_UNKNOWNCMDID = 0x8
    """Unknown command ID."""
    ERR_BADTASKID = 0x9
    """Invalid task ID."""
    ERR_NOIO = 0xA
    """No IO."""
    ERR_UNKNOWNAMSCMD = 0xB
    """Unknown AMS command."""
    ERR_WIN32ERROR = 0xC
    """Win32 error."""
    ERR_PORTNOTCONNECTED = 0xD
    """Port not connected."""
    ERR_INVALIDAMSLENGTH = 0xE
    """Invalid AMS length."""
    ERR_INVALIDAMSNETID = 0xF
    """Invalid AMS Net ID."""
    ERR_LOWINSTLEVEL = 0x10
    """Installation level is too low - TwinCAT 2 license error."""
    ERR_NODEBUGINTAVAILABLE = 0x11
    """No debugging available."""
    ERR_PORTDISABLED = 0x12
    """Port disabled - TwinCAT system service not started."""
    ERR_PORTALREADYCONNECTED = 0x13
    """Port already connected."""
    ERR_AMSSYNC_W32ERROR = 0x14
    """AMS Sync Win32 error."""
    ERR_AMSSYNC_TIMEOUT = 0x15
    """AMS Sync Timeout."""
    ERR_AMSSYNC_AMSERROR = 0x16
    """AMS Sync error."""
    ERR_AMSSYNC_NOINDEXINMAP = 0x17
    """No index map for AMS Sync available."""
    ERR_INVALIDAMSPORT = 0x18
    """Invalid AMS port."""
    ERR_NOMEMORY = 0x19
    """No memory."""
    ERR_TCPSEND = 0x1A
    """TCP send error."""
    ERR_HOSTUNREACHABLE = 0x1B
    """Host unreachable."""
    ERR_INVALIDAMSFRAGMENT = 0x1C
    """Invalid AMS fragment."""
    ERR_TLSSEND = 0x1D
    """TLS send error - secure ADS connection failed."""
    ERR_ACCESSDENIED = 0x1E
    """Access denied - secure ADS access denied."""

    ROUTERERR_NOLOCKEDMEMORY = 0x500
    """Locked memory cannot be allocated."""
    ROUTERERR_RESIZEMEMORY = 0x501
    """The router memory size could not be changed."""
    ROUTERERR_MAILBOXFULL = 0x502
    """The mailbox has reached the maximum number of possible messages."""
    ROUTERERR_DEBUGBOXFULL = 0x503
    """The Debug mailbox has reached the maximum number of possible messages."""
    ROUTERERR_UNKNOWNPORTTYPE = 0x504
    """The port type is unknown."""
    ROUTERERR_NOTINITIALIZED = 0x505
    """The router is not initialized."""
    ROUTERERR_PORTALREADYINUSE = 0x506
    """The port number is already assigned."""
    ROUTERERR_NOTREGISTERED = 0x507
    """The port is not registered."""
    ROUTERERR_NOMOREQUEUES = 0x508
    """The maximum number of ports has been reached."""
    ROUTERERR_INVALIDPORT = 0x509
    """The port is invalid."""
    ROUTERERR_NOTACTIVATED = 0x50A
    """The router is not active."""
    ROUTERERR_FRAGMENTBOXFULL = 0x50B
    """The mailbox has reached the maximum number for fragmented messages."""
    ROUTERERR_FRAGMENTTIMEOUT = 0x50C
    """A fragment timeout has occurred."""
    ROUTERERR_TOBEREMOVED = 0x50D
    """The port is removed."""

    ADSERR_DEVICE_ERROR = 0x0700
    """General device error."""
    ADSERR_DEVICE_SRVNOTSUPP = 0x0701
    """Service is not supported by the server."""
    ADSERR_DEVICE_INVALIDGRP = 0x0702
    """Invalid index group."""
    ADSERR_DEVICE_INVALIDOFFSET = 0x0703
    """Invalid index offset."""
    ADSERR_DEVICE_INVALIDACCESS = 0x0704
    """Reading or writing not permitted."""
    ADSERR_DEVICE_INVALIDSIZE = 0x0705
    """Parameter size not correct."""
    ADSERR_DEVICE_INVALIDDATA = 0x0706
    """Invalid data values."""
    ADSERR_DEVICE_NOTREADY = 0x0707
    """Device is not ready to operate."""
    ADSERR_DEVICE_BUSY = 0x0708
    """Device is busy."""
    ADSERR_DEVICE_INVALIDCONTEXT = 0x0709
    """Invalid operating system context.
    This can result from use of ADS blocks in different tasks.
    (possible resolution includes multitasking synchronization in the PLC)"""
    ADSERR_DEVICE_NOMEMORY = 0x070A
    """Insufficient memory."""
    ADSERR_DEVICE_INVALIDPARM = 0x070B
    """Invalid parameter values."""
    ADSERR_DEVICE_NOTFOUND = 0x070C
    """Not found (files, ...)."""
    ADSERR_DEVICE_SYNTAX = 0x070D
    """Syntax error in file or command."""
    ADSERR_DEVICE_INCOMPATIBLE = 0x070E
    """Objects do not match."""
    ADSERR_DEVICE_EXISTS = 0x070F
    """Object already exists."""
    ADSERR_DEVICE_SYMBOLNOTFOUND = 0x0710
    """Symbol not found."""
    ADSERR_DEVICE_SYMBOLVERSIONINVALID = 0x0711
    """Invalid symbol version.
    This can occur due to an online change. Create a new handle."""
    ADSERR_DEVICE_INVALIDSTATE = 0x0712
    """Device (server) is in invalid state."""
    ADSERR_DEVICE_TRANSMODENOTSUPP = 0x0713
    """AdsTransMode not supported."""
    ADSERR_DEVICE_NOTIFYHNDLINVALID = 0x0714
    """Notification handle is invalid."""
    ADSERR_DEVICE_CLIENTUNKNOWN = 0x0715
    """Notification client not registered."""
    ADSERR_DEVICE_NOMOREHDLS = 0x0716
    """No further handle available."""
    ADSERR_DEVICE_INVALIDWATCHSIZE = 0x0717
    """Notification size too large."""
    ADSERR_DEVICE_NOTINIT = 0x0718
    """Device not initialized."""
    ADSERR_DEVICE_TIMEOUT = 0x0719
    """Device has a timeout."""
    ADSERR_DEVICE_NOINTERFACE = 0x071A
    """Interface query failed."""
    ADSERR_DEVICE_INVALIDINTERFACE = 0x071B
    """Wrong interface requested."""
    ADSERR_DEVICE_INVALIDCLSID = 0x071C
    """Class ID is invalid."""
    ADSERR_DEVICE_INVALIDOBJID = 0x071D
    """Object ID is invalid."""
    ADSERR_DEVICE_PENDING = 0x071E
    """Request pending."""
    ADSERR_DEVICE_ABORTED = 0x071F
    """Request is aborted."""
    ADSERR_DEVICE_WARNING = 0x0720
    """Signal warning."""
    ADSERR_DEVICE_INVALIDARRAYIDX = 0x0721
    """Invalid array index."""
    ADSERR_DEVICE_SYMBOLNOTACTIVE = 0x0722
    """Symbol not active."""
    ADSERR_DEVICE_ACCESSDENIED = 0x0723
    """Access denied."""
    ADSERR_DEVICE_LICENSENOTFOUND = 0x0724
    """Missing license."""
    ADSERR_DEVICE_LICENSEEXPIRED = 0x0725
    """License expired."""
    ADSERR_DEVICE_LICENSEEXCEEDED = 0x0726
    """License exceeded."""
    ADSERR_DEVICE_LICENSEINVALID = 0x0727
    """Invalid license."""
    ADSERR_DEVICE_LICENSESYSTEMID = 0x0728
    """License problem: System ID is invalid."""
    ADSERR_DEVICE_LICENSENOTIMELIMIT = 0x0729
    """License not limited in time."""
    ADSERR_DEVICE_LICENSEFUTUREISSUE = 0x072A
    """Licensing problem: time in the future."""
    ADSERR_DEVICE_LICENSETIMETOLONG = 0x072B
    """License period too long."""
    ADSERR_DEVICE_EXCEPTION = 0x072C
    """Exception at system startup."""
    ADSERR_DEVICE_LICENSEDUPLICATED = 0x072D
    """License file read twice."""
    ADSERR_DEVICE_SIGNATUREINVALID = 0x072E
    """Invalid signature."""
    ADSERR_DEVICE_CERTIFICATEINVALID = 0x072F
    """Invalid certificate."""
    ADSERR_DEVICE_LICENSEOEMNOTFOUND = 0x0730
    """Public key not known from OEM."""
    ADSERR_DEVICE_LICENSERESTRICTED = 0x0731
    """License not valid for this system ID."""
    ADSERR_DEVICE_LICENSEDEMODENIED = 0x0732
    """Demo license prohibited."""
    ADSERR_DEVICE_INVALIDFNCID = 0x0733
    """Invalid function ID."""
    ADSERR_DEVICE_OUTOFRANGE = 0x0734
    """Outside the valid range."""
    ADSERR_DEVICE_INVALIDALIGNMENT = 0x0735
    """Invalid alignment."""
    ADSERR_DEVICE_LICENSEPLATFORM = 0x0736
    """Invalid platform level."""
    ADSERR_DEVICE_FORWARD_PL = 0x0737
    """Context - forward to passive level."""
    ADSERR_DEVICE_FORWARD_DL = 0x0738
    """Context - forward to dispatch level."""
    ADSERR_DEVICE_FORWARD_RT = 0x0739
    """Context - forward to real time."""
    ADSERR_CLIENT_ERROR = 0x0740
    """Client error."""
    ADSERR_CLIENT_INVALIDPARM = 0x0741
    """Service contains an invalid parameter."""
    ADSERR_CLIENT_LISTEMPTY = 0x0742
    """Polling list is empty."""
    ADSERR_CLIENT_VARUSED = 0x0743
    """Var connection already in use."""
    ADSERR_CLIENT_DUPLINVOKEID = 0x0744
    """The called ID is already in use."""
    ADSERR_CLIENT_SYNCTIMEOUT = 0x0745
    """Timeout has occurred - no response from the remote terminal in the ADS timeout.
    The route setting of the remote terminal may be configured incorrectly."""
    ADSERR_CLIENT_W32ERROR = 0x0746
    """Error in Win32 subsystem."""
    ADSERR_CLIENT_TIMEOUTINVALID = 0x0747
    """Invalid client timeout value."""
    ADSERR_CLIENT_PORTNOTOPEN = 0x0748
    """Port not open."""
    ADSERR_CLIENT_NOAMSADDR = 0x0749
    """No AMS address."""
    ADSERR_CLIENT_SYNCINTERNAL = 0x0750
    """Internal error in Ads sync."""
    ADSERR_CLIENT_ADDHASH = 0x0751
    """Hash table overflow."""
    ADSERR_CLIENT_REMOVEHASH = 0x0752
    """Key not found in the table."""
    ADSERR_CLIENT_NOMORESYM = 0x0753
    """No symbols in the cache."""
    ADSERR_CLIENT_SYNCRESINVALID = 0x0754
    """Invalid response received."""
    ADSERR_CLIENT_SYNCPORTLOCKED = 0x0755
    """Sync Port is locked."""
    ADSERR_CLIENT_REQUESTCANCELLED = 0x0756
    """The request was cancelled."""

    RTERR_INTERNAL = 0x1000
    """Internal error in the real-time system."""
    RTERR_BADTIMERPERIODS = 0x1001
    """Timer value is not valid."""
    RTERR_INVALIDTASKPTR = 0x1002
    """Task pointer has the invalid value 0 (zero)."""
    RTERR_INVALIDSTACKPTR = 0x1003
    """Stack pointer has the invalid value 0 (zero)."""
    RTERR_PRIOEXISTS = 0x1004
    """The request task priority is already assigned."""
    RTERR_NOMORETCB = 0x1005
    """No free TCB (Task Control Block) available. The maximum number of TCBs is 64."""
    RTERR_NOMORESEMAS = 0x1006
    """No free semaphores available. The maximum number of semaphores is 64."""
    RTERR_NOMOREQUEUES = 0x1007
    """No free space available in the queue.
    The maximum number of positions in the queue is 64."""
    RTERR_EXTIRQALREADYDEF = 0x100D
    """An external synchronization interrupt is already applied."""
    RTERR_EXTIRQNOTDEF = 0x100E
    """No external sync interrupt applied."""
    RTERR_EXTIRQINSTALLFAILED = 0x100F
    """Application of the external synchronization interrupt has failed."""
    RTERR_IRQLNOTLESSOREQUAL = 0x1010
    """Call of a service function in the wrong context"""
    RTERR_VMXNOTSUPPORTED = 0x1017
    """Intel VT-x extension is not supported."""
    RTERR_VMXDISABLED = 0x1018
    """Intel VT-x extension is not enabled in the BIOS."""
    RTERR_VMXCONTROLSMISSING = 0x1019
    """Missing function in Intel VT-x extension."""
    RTERR_VMXENABLEFAILS = 0x101A
    """Activation of Intel VT-x fails."""



from __future__ import annotations

import inspect
import logging
import re
import socket
from collections.abc import Callable

import numpy as np
import numpy.typing as npt

from ._constants import TWINCAT_STRING_ENCODING


def get_localhost_name() -> str:
    """
    Get the hostname of the local machine.

    :returns: the local machine hostname
    """
    return socket.gethostname()


def get_localhost_ip() -> str:
    """
    Get the IP address of the local machine.

    :returns: the local machine IP address
    """
    return socket.gethostbyname(get_localhost_name())


def get_local_netid_str() -> str:
    """
    Create the ams netid string value of the Ads client (localhost).

    :returns: the string representing the local client netid
    """
    return get_localhost_ip() + ".1.1"


def bytes_to_string(raw_data: bytes, strip: bool = True) -> str:
    """
    Convert a bytes object into a unicode string.

    :param raw_data: an array of bytes to convert to string
    :param strip: boolean indicating whether trailing bytes must be removed

    :returns: a string object as a numpy string type
    """
    if strip:
        null_index = raw_data.find(0)
        if null_index != -1:
            raw_data = raw_data[:null_index]
    return raw_data.decode(encoding=TWINCAT_STRING_ENCODING)


def add_comment(new_str: str, old_str: str) -> str:
    """
    Concatenate two strings with a new line.

    :param new_str: new string to append
    :param old_str: existing string to append to

    :returns: an updated string
    """
    return new_str if not old_str else "\n".join([old_str, new_str])


def process_notifications(
    func: Callable,
    notifications: npt.NDArray,
) -> npt.NDArray:
    """
    Manipulate the received notification array by applying a given function.
    This method may be used to test the load on the client resources.

    :param func: the processing function to apply to the notification data
    :param notifications: a numpy array comprising multiple ADS notifications

    :returns: the post-processed notification array
    """
    args = inspect.getfullargspec(func).args
    assert len(args) == 1, (
        f"The processing function {func.__name__} takes more than 1 argument."
    )
    input = inspect.signature(func).parameters[args[0]]
    assert input.annotation == "np.ndarray", (
        f"The processing function {func.__name__} requires a numpy array as argument."
    )
    data = func(notifications)
    logging.debug(
        f"Applied '{func.__name__}' function "
        + f"to notification data comprising {len(data[0])} fields"
    )
    return data


def average(array: np.ndarray) -> np.ndarray:
    """
    Average data from all fields in a numpy structured array.

    :param array: a numpy structured array comprising multiple fields

    :returns: a 1D numpy array with averaged values
    """
    mean_array = np.empty(1, dtype=array.dtype)
    assert array.dtype.fields is not None
    for field in array.dtype.fields:
        mean_array[field] = np.mean(array[field])
    return mean_array


def get_notification_changes(
    new_array: np.ndarray, old_array: np.ndarray
) -> np.ndarray:
    """
    Compare two notification arrays and return the differences.
    Each array is expected to be a numpy structured arra.
    Both arrays must have the same shape and dtype.

    :param new_array: the new structured array with updated values
    :param old_array: the old structured array to compare against

    :returns: a numpy array containing the differences between the two arrays
    """
    assert new_array.shape == old_array.shape
    assert new_array.dtype == old_array.dtype
    assert new_array[0].size == old_array[0].size

    mask = []
    for a, b in zip(new_array[0], old_array[0], strict=True):
        mask.append(not np.array_equal(a, b) if isinstance(a, np.ndarray) else a != b)

    diff = []
    assert new_array.dtype.names
    for val, name in zip(mask, new_array.dtype.names, strict=True):
        if val:
            diff.append(name)
    return new_array[diff]


def filetime_to_dt(filetime: int) -> np.datetime64:
    """
    Convert a Windows FILETIME timestamp to a numpy datetime64 object.
    FILETIME is in 100-nanosecond intervals since January 1, 1601 (UTC).
    Numpy datetime64 is in nanoseconds since January 1, 1970 (UTC).

    :param filetime: the FILETIME timestamp as a 64-bit integer

    :returns: the corresponding numpy datetime64 object
    """
    # Difference between epochs in 100-nanosecond intervals
    epoch_diff = 116444736000000000
    # Number of 100-nanosecond intervals in a second
    hundred_nanoseconds = 10_000_000

    # Convert FILETIME to seconds since Unix epoch
    unix_time = (filetime - epoch_diff) / hundred_nanoseconds

    # Convert to numpy datetime64 inc. seconds to nanoseconds conversion
    return np.datetime64(int(unix_time * 1e9), "ns")


def trim_ecat_name(name: str) -> str:
    """
    Shorten and remove spaces from the original EtherCAT name.

    :param name: the original EtherCAT device/terminal name

    :returns: a trimmed name without spaces
    """
    matches = re.search(r"^(\w+\s+)\d+", name)
    return matches.group(0).replace(" ", "") if matches else name


def check_ndarray(
    obj: npt.NDArray,
    expected_dtype: npt.DTypeLike,
    expected_shape: tuple[int,] | tuple[int, int],
) -> bool:
    """
    Check if an object is a numpy ndarray and verifies its dtype and shape.

    :param obj:
    :param expected_dtype:
    :param expected_shape:

    :returns: true if the object is a numpy array with the expected dtype and shape
    """
    return (
        isinstance(obj, np.ndarray)
        and obj.dtype == expected_dtype
        and obj.shape == expected_shape
    )



import asyncio

from fastcs_catio.client import AsyncioADSClient

# Connection to CX2020 device
TARGET_IP = "172.23.240.142"
TARGET_NETID = "5.59.238.150.1.1"
TARGET_ADS_PORT = 27909


async def make_client() -> AsyncioADSClient:
    return await AsyncioADSClient.connected_to(
        target_ip=TARGET_IP,
        target_ams_net_id=TARGET_NETID,
        target_ams_port=TARGET_ADS_PORT,
    )


async def main():
    client = await make_client()

    try:
        await client.introspect_io_server()

        # Get a slave terminal of a given type
        slave_type = "EL3104"
        terminal = client.find_slave_in_master_device(slave_type)

        # Toggle one of its CoE parameter, e.g. the 'filter enable'
        if terminal:
            await client.set_io_coe_parameter(terminal, "8000", "0006", True)
            await asyncio.sleep(2)
            await client.set_io_coe_parameter(terminal, "8000", "0006", False)
        else:
            print(
                f"ERROR: no slave terminal of type {slave_type} is available \
                    on the EtherCAT Master device."
            )

    finally:
        await client.close()


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import datetime
import inspect
import random

from fastcs_catio.client import AsyncioADSClient

# Connection to CX2020 device
TARGET_IP = "172.23.240.142"
TARGET_NETID = "5.59.238.150.1.1"
TARGET_ADS_PORT = 27909


async def query_something(client: AsyncioADSClient) -> None:
    print("...checking status from random slaves...")

    try:
        device = next(iter(client._ecdevices.values()))
        for _ in range(5):
            await client.check_slave_states(
                device.id, random.choice(device.slaves).address
            )
            await client.check_slave_crc(
                device.id, random.choice(device.slaves).address
            )
            await asyncio.sleep(1)

    except ValueError as err:
        print(f"Value error raised in function '{inspect.stack()[0][3]}'")
        raise ValueError from err
    except AssertionError:
        print(f"Assertion error raised in function '{inspect.stack()[0][3]}'")
        raise

    print("...query task completed...")


async def poll_continuously(client: AsyncioADSClient, duration: float) -> None:
    print("...continuous polling...")
    loop = asyncio.get_running_loop()
    end_time = loop.time() + duration
    sleep = 1.0
    await client.reset_frame_counters()
    while True:
        print(f"{datetime.datetime.now()} ...polling")
        if (loop.time() + sleep) >= end_time:
            break
        try:
            await client.poll_states()
            await client.poll_crc_counters()
            await client.poll_frame_counters()
        except ValueError as err:
            print(f"Value error raised in function '{inspect.stack()[0][3]}'")
            raise ValueError from err
        except AssertionError:
            print(f"Assertion error raised in function '{inspect.stack()[0][3]}'")
            raise
        await asyncio.sleep(sleep)
    print("...polling task completed...")


async def make_client() -> AsyncioADSClient:
    return await AsyncioADSClient.connected_to(
        target_ip=TARGET_IP,
        target_ams_net_id=TARGET_NETID,
        target_ams_port=TARGET_ADS_PORT,
    )


async def main():
    client = await make_client()

    try:
        await client.introspect_io_server()

        async with asyncio.TaskGroup() as tg:
            task0 = tg.create_task(
                coro=query_something(client), name="Query_task", context=None
            )
            task1 = tg.create_task(
                coro=poll_continuously(client, 12.0), name="Poll_task", context=None
            )

        print(f"All tasks have completed now: {task0.result()}, {task1.result()}")

    finally:
        await client.close()


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import logging
import time
from collections.abc import Sequence
from dataclasses import dataclass, field
from typing import Any, Self, SupportsInt

import numpy.typing as npt
from fastcs.tracer import Tracer

from fastcs_catio.devices import AdsSymbol

from .client import AsyncioADSClient


class CATioFastCSRequest:
    """
    Request object sent to the catio client (string subclass).
    Used to encapsulate all the information needed to perform a query.
    """

    def __init__(self, command: str, *args, **kwargs):
        self.command = command
        """The command to be executed by the CATio client."""
        self.args = args
        """Optional positional arguments for the command."""
        self.kwargs = kwargs
        """Optional keyword arguments for the command."""

    def __repr__(self) -> str:
        """Return a string representation of the CATio request."""
        return repr(
            self.command
            + "("
            + ", ".join(self.args)
            + ", "
            + ", ".join([f"{k}={v!r}" for k, v in self.kwargs.items()])
            + ")"
        )


@dataclass
class CATioFastCSResponse:
    """
    Response object received by the catio client.
    Used to potentially manipulate the received information.
    """

    value: Any
    """The response received from the CATio client."""

    def to_string(self) -> str:
        """Attempt to return a string representation of the CATio response.

        :returns: the CATio response to a query as a string

        :raises ValueError: if the received response cannot be stringified
        """
        try:
            return repr(self.value)
        except Exception as err:
            raise ValueError from err


class DisconnectedError(Exception):
    """Raised if the IP connection is disconnected."""

    pass


@dataclass
class CATioServerConnectionSettings:
    """
    Settings required to establish a TCP connection with a CATio server.
    Act as a wrapper for connection parameters.
    """

    ip: str = "127.0.0.1"
    """The IP address of the TwinCAT server to connect to."""
    ams_netid: str = "127.0.0.1.1.1"
    """The Ams netid of the TwinCAT server to connect to."""
    ams_port: int = 25565
    """The Ams port of the TwinCAT server to connect to."""

    def __repr__(self) -> str:
        return f"TCP connection to remote server with netid {self.ams_netid}, \
            at address {self.ip}, on port {self.ams_port}"


@dataclass
class CATioStreamConnection:
    """
    For setting up a CATio client able to read and write to a stream.
    Act as a wrapper for interacting with an AsyncioADSClient \
        and handling I/O communications.
    One instance of this class should be created per TCP connection.
    One instance of AsyncioADSClient will be created and managed internally.
    """

    _connection_settings: CATioServerConnectionSettings
    """The connection settings used to connect to the CATio server."""
    _catio_client: AsyncioADSClient
    """The ADS client used to communicate with the CATio server."""
    _notification_symbols: dict[SupportsInt, Sequence[AdsSymbol]] = field(
        default_factory=dict
    )
    """A mapping of device ids to their corresponding notification symbols."""
    _subscribed_symbols: list[AdsSymbol] = field(default_factory=list)
    """The list of currently subscribed notification symbols."""

    @property
    def settings(self) -> CATioServerConnectionSettings:
        """The connection settings used to connect to the CATio server."""
        return self._connection_settings

    @property
    def client(self) -> AsyncioADSClient:
        """The ADS client used to communicate with the CATio server."""
        return self._catio_client

    @property
    def notification_symbols(self) -> dict[SupportsInt, Sequence[AdsSymbol]]:
        """A mapping of device ids to their corresponding notification symbols."""
        return self._notification_symbols

    @property
    def subscribed_symbols(self) -> list[AdsSymbol]:
        """The list of currently subscribed notification symbols."""
        return self._subscribed_symbols

    @classmethod
    async def connect(cls, settings: CATioServerConnectionSettings) -> Self:
        """
        Create a client which will connect to the TwinCAT server and \
            support ADS communication with the attached I/O devices.

        :param settings: the connection settings to use for connecting to the server

        :returns: an instance of CATioStreamConnection with an active client
        """
        ads_client = await AsyncioADSClient.connected_to(
            target_ip=settings.ip,
            target_ams_net_id=settings.ams_netid,
            target_ams_port=settings.ams_port,
        )
        return cls(settings, ads_client)

    def __post_init__(self) -> None:
        """
        Initialise the asyncio lock for managing concurrent access to the client.
        """
        self._lock = asyncio.Lock()

    async def __aenter__(self) -> Self:
        """
        Acquire the asyncio lock to ensure exclusive access to the client.
        """
        await self._lock.acquire()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """
        Release the asyncio lock to allow other coroutines to access the client.
        """
        self._lock.release()
        # pass

    async def initialise(self) -> None:
        """
        Update the ads client with the current I/O server configuration.
        This includes the detection of all configured hardware in the EtherCAT system
        and of all accessible ads symbol variables.

        If the server configuration changes, this method should be called again.
        """
        await self.client.introspect_io_server()
        self._notification_symbols = await self.client.get_all_symbols()

    async def command(self, command: str, *args, **kwargs) -> None:
        """
        Send a message command to the CATio client.
        This message command will be translated as an API function call.

        :param command: the command to send to the client
        :param args: positional arguments for the command
        :param kwargs: keyword arguments for the command
        """
        try:
            await self.client.command(command, *args, **kwargs)
        except ValueError as err:
            logging.debug(f"API call failed with error: {err}")

    async def query(self, message: CATioFastCSRequest) -> CATioFastCSResponse:
        """
        Send a message request to the catio client.
        This message request will be translated as an API function call.

        :param message: a CATio request message which will be routed via the client.

        :returns: the response to the query as received by the CATio client
        """
        response = ""
        try:
            response = await self.client.query(
                message.command, *message.args, **message.kwargs
            )
        except ValueError as err:
            logging.debug(f"API call failed with error: {err}")

        # # Very verbose logging!
        # logging.debug(f"CATio client response to '{message}' query: {response}")
        return CATioFastCSResponse(response)

    async def add_notifications(self, device_id: int) -> None:
        """
        Register symbol notifications with the ads client for a given device.
        This will include all ads symbols available to this device.

        :param device_id: the id of the EtherCAT device to subscribe to

        :raises ValueError: if no notification symbols are found for the device
        """
        if device_id not in self.notification_symbols:
            raise ValueError(
                f"No notification symbols found for device id {device_id}."
            )
        subscription_symbols = self.notification_symbols[device_id]
        self._subscribed_symbols = list(subscription_symbols)

        await self.client.add_notifications(
            subscription_symbols,  # max_delay_ms=1000, cycle_time_ms=1000
        )
        logging.info(
            f"Subscribed to {len(subscription_symbols)} symbols "
            + f"for device id {device_id}."
        )
        # Small delay to avoid overloading the server after the subscription process.
        await asyncio.sleep(0.3)

    def monitor_notifications(self, enabled: bool, flush_period: float = 0.5) -> None:
        """
        Enable or disable the periodic monitoring of symbol notifications by the client.
        This will start/stop a background task which periodically checks for new
        notifications and processes them.

        :param enabled: True to enable notification monitoring, False to disable it
        :param flush_period: the period (in seconds) at which notifications are flushed
        """
        if enabled:
            self.client.start_notification_monitor(flush_period)
        else:
            self.client.stop_notification_monitor()

    async def get_notifications(self, timeout: int = 60) -> npt.NDArray:
        """
        Get the latest ads symbol notifications from the ads client.
        This will wait for new notifications to arrive, up to the specified timeout.

        :param timeout: the maximum time to wait for notifications (in seconds)

        :returns: a numpy array containing the latest notifications
        """
        return await self.client.get_notifications(timeout)

    async def delete_all_notifications(self) -> None:
        """
        Delete the existing ads symbol notifications.
        This will unsubscribe from all currently subscribed symbols.

        :param device_id: the id of the EtherCAT device to subscribe to
        """
        logging.info("...deleting active notifications...")
        await self.client.delete_notifications(self._subscribed_symbols)

    async def close(self) -> None:
        """
        Stop any background tasks related to notification monitoring, \
            unsubscribe to all symbol notifications \
                and close the client connection with the catio server.
        """
        await self.delete_all_notifications()
        await asyncio.sleep(1)

        await self.client.close()


class CATioConnection(Tracer):
    """
    For connecting to a Beckhoff TwinCAT server using a TCP connection.
    This class manages a CATioStreamConnection instance \
        which handles the underlying ADS communication.
    One instance of this class should be created per TCP connection.
    """

    def __init__(self, connection: CATioStreamConnection | None = None):
        super().__init__()
        self.__connection: CATioStreamConnection | None = connection
        """The underlying CATio stream connection."""

    @property
    def _connection(self) -> CATioStreamConnection:
        """
        The underlying CATio stream connection.

        :raises DisconnectedError: if no connection is established
        """
        if self.__connection is None:
            raise DisconnectedError(
                "No open connection with the CATio system. Call connect() first."
            )
        return self.__connection

    @_connection.setter
    def _connection(self, value: CATioStreamConnection | None) -> None:
        """
        Set the underlying CATio stream connection.

        :param value: the new connection to set, or None to disconnect
        """
        self.__connection = value

    @property
    def settings(self) -> CATioServerConnectionSettings:
        """The connection settings used to connect to the CATio server."""
        return self._connection.settings

    @property
    def client(self) -> AsyncioADSClient:
        """The ADS client used to communicate with the CATio server."""
        return self._connection.client

    def is_defined(self) -> bool:
        """
        Check if a valid CATio connection has been defined.

        :returns: True if a connection is established, False otherwise
        """
        return self.__connection is not None

    async def connect(self, settings: CATioServerConnectionSettings) -> None:
        """
        Establish a TCP connection and enable stream communication.

        :param settings: the connection settings to use for connecting to the server
        """
        self._connection = await CATioStreamConnection.connect(settings)
        logging.info(
            f"Opened stream communication with ADS server at {time.strftime('%X')}"
        )

    async def initialise(self) -> None:
        """
        Initialise the client connection with the current server settings.
        This includes the detection of all configured hardware in the EtherCAT system \
            and of all accessible ads symbol variables.
        """
        await self._connection.initialise()

    async def send_command(self, message: CATioFastCSRequest) -> None:
        """
        Send a message command to the CATio client.
        This message command will be translated as an API function call to the client.

        :param message: a CATio request message which will be routed via the client
        """
        async with self._connection as connection:
            await connection.command(message.command, *message.args, **message.kwargs)

    async def send_query(self, message: CATioFastCSRequest) -> Any:
        """
        Send a message request to the CATio client and return the response.
        This message request will be translated as an API function call to the client.

        param message: a CATio request message which will be routed via the client.

        :returns: the response to the query as received by the CATio client
        """
        async with self._connection as connection:
            response = await connection.query(message)
            self.log_event(
                "Received query response",
                query=message.command,
                response=response.to_string(),
            )
            return response.value

    async def close(self) -> None:
        """
        Stop the communication stream and \
            close the TCP connection with the TwinCAT server.
        """
        async with self._connection as connection:
            await connection.close()
            self._connection = None
        logging.info(
            f"Closed stream communication with ADS server at {time.strftime('%X')}"
        )

    async def add_notifications(self, device_id: int) -> None:
        """
        Add symbol notifications for a given EtherCAT device on the I/O server.
        This will include all ads symbols available to this device.

        :param device_id: the id of the device whose notifications must be setup
        """
        await self._connection.add_notifications(device_id)

    async def get_notification_streams(self, timeout: int = 60) -> npt.NDArray:
        """
        Get the latest ads symbol notifications from the connection stream.
        This will wait for new notifications to arrive, up to the specified timeout.

        :param timeout: the maximum time to wait for notifications (in seconds)

        :returns: a numpy array containing the latest notifications
        """
        return await self._connection.get_notifications(timeout)

    def enable_notification_monitoring(
        self, enabled: bool, flush_period: float = 0.5
    ) -> None:
        """
        Enable or disable the periodic monitoring of notifications.
        This will start/stop a background task which periodically checks for \
            new notifications and processes them.

        :param enabled: True to enable notification monitoring, False to disable it
        :param flush_period: the period (in seconds) at which notifications are flushed
        """
        if enabled:
            self._connection.monitor_notifications(True, flush_period)
            logging.debug("Notification monitoring enabled.")
        else:
            self._connection.monitor_notifications(False)
            logging.debug("Notification monitoring disabled.")



from collections import namedtuple
from collections.abc import Generator, Sequence
from dataclasses import dataclass
from enum import Enum
from typing import Any, Self, SupportsInt

import numpy as np
import numpy.typing as npt

from ._constants import (
    AdsDataType,
    DeviceType,
    SymbolFlag,
)
from ._types import AmsNetId
from .messages import DeviceFrames, IOIdentity, SlaveCRC, SlaveState

ChainLocation = namedtuple("ChainLocation", ["node", "position"])

STD_UPDATE_POLL_PERIOD: float = 2.0
FAST_UPDATE_POLL_PERIOD: float = 0.2
NOTIF_UPDATE_POLL_PERIOD: float = 1.0

OVERSAMPLING_FACTOR = 100
ELM_OVERSAMPLING_FACTOR = 50


# ===================================================================
# ===== EtherCAT OBJECTS
# ===================================================================


class IONodeType(str, Enum):
    """
    Provide a broad classification of the type of hardware found in the EtherCAT system.
    """

    Server = "server"
    Device = "device"
    Coupler = "coupler"
    Slave = "slave"


@dataclass
class AdsSymbol:
    """
    Define an ADS symbol.
    """

    parent_id: SupportsInt
    """Id of the device which the symbol belongs to"""
    name: str
    """Name of the symbol"""
    dtype: npt.DTypeLike
    """Data type of the symbol"""
    size: int
    """Number of elements"""
    group: SupportsInt
    """Index group used by the ADS protocol to address the symbol"""
    offset: SupportsInt
    """Index offset used by the ADS protocol to address the symbol"""
    comment: str
    """Optional comment associated to the symbol"""
    handle: SupportsInt | None = None
    """Unique handle value mapping the symbol to an ADS notification"""

    @property
    def datatype(self) -> npt.DTypeLike:
        """
        Get the numpy data type of the data associated to the symbol value.
        It takes into account the type and size of the symbol, i.e. it may return:
        - a generic type: e.g. uint16 of type <class 'numpy.dtypes.UInt16DType'>
        - an array type: e.g. ('<i2', (100,)) of type <class 'numpy.dtypes.VoidDType'>

        :returns: the extended data type of the ADS symbol value.
        """
        if self.size > 1:
            return np.dtype((self.dtype, self.size))
        return np.dtype(self.dtype)

    @property
    def nbytes(self) -> int:
        """
        Get the total number of bytes of the data associated to the symbol value.

        :returns: the total size in bytes of the ADS symbol value
        """
        return np.dtype(self.dtype).itemsize * self.size


@dataclass
class AdsSymbolNode:
    """
    Define a distinct symbol node as exposed by the uploaded symbol table.
    """

    parent_id: SupportsInt
    """Id of the device which the symbol node belongs to"""
    name: str
    """Name of the symbol node (i.e. root name of the device symbol)"""
    type_name: str
    """Type of the symbol node as characterised by the generic terminal type
    (i.e. not the actual data type of the symbol)."""
    ads_type: AdsDataType
    """Actual data type of the symbol"""
    size: SupportsInt
    """Size of the symbol in bytes (0 corresponds to 'bit')"""
    index_group: SupportsInt
    """Index group used by the ADS protocol to address the symbol node"""
    index_offset: SupportsInt
    """Index offset used by the ADS protocol to address the symbol node"""
    flag: SymbolFlag
    """ADS flag characterising the symbol node"""
    comment: str
    """Optional comment associated to the symbol node"""


@dataclass
class IOSlave:
    """
    Define an EtherCAT slave object configured on an EtherCAT device.
    """

    parent_device: int
    """The id of the EtherCAT device the slave belongs to"""
    type: str
    """The CANopen type object of the slave"""
    name: str
    """The CANopen name object of the slave"""
    address: SupportsInt
    """The EtherCAT address of the slave"""
    identity: IOIdentity
    """The CANopen identity object of the slave"""
    states: SlaveState
    """The EtherCAT states of the slave"""
    crcs: SlaveCRC
    """The error counters for the slave CRC on the distinct communication ports"""
    crc_error_sum: SupportsInt = 0
    """The error sum counter for the cyclic redundancy check of the slave"""
    loc_in_chain: ChainLocation = ChainLocation(0, 0)
    """The position of the slave within the EtherCAT device chain"""
    category: IONodeType = IONodeType.Slave
    """The component category the object belongs to in the EtherCAT system"""

    def get_type_name(self) -> str:
        """
        Translate the Beckhoff terminal type name into a more suitable PV name.
        """
        if self.category == "coupler":
            return f"RIO{self.loc_in_chain.node}"
        elif self.category == "slave":
            # This name could be updated by the actual Terminal Class (ai,ao,di,do...)?
            return f"MOD{self.loc_in_chain.position}"
        else:
            raise NameError(f"I/O terminal category '{self.category}' isn't valid.")


@dataclass
class IODevice:
    """
    Define an EtherCAT device object registered on the I/O server.
    """

    id: SupportsInt
    """The id number associated with the EtherCAT device"""
    type: DeviceType
    """The type of the EtherCAT device"""
    name: str
    """The name of the EtherCAT device"""
    netid: AmsNetId
    """The ams netid address of the EtherCAT device"""
    identity: IOIdentity
    """The CANopen identity object of the EtherCAT device"""
    frame_counters: DeviceFrames
    """The EtherCAT cycle frame counters for the EtherCAT device"""
    slave_count: SupportsInt
    """The number of slave terminals configured on the EtherCAT device"""
    slaves_states: Sequence[SlaveState]
    """The states values for all of the slaves"""
    slaves_crc_counters: Sequence[np.uint32]
    """The error sum counter values of the cyclic redundancy check for all slaves"""
    slaves: Sequence[IOSlave]
    """The slave terminals configured on the EtherCAT device"""
    node_count: SupportsInt = 0
    """The number of node terminals (i.e. couplers) among the configured slaves"""
    category: IONodeType = IONodeType.Device
    """The component category the object belongs to in the EtherCAT system"""

    def __repr__(self) -> str:
        return (
            f"IODevice(id={self.id}, type={self.type}, name={self.name}, "
            + f"netid={self.netid}, slaveCount={self.slave_count}, "
            + f"slaveAdresses=[{self.slaves[0].address}...{self.slaves[-1].address}])"
        )

    def get_type_name(self) -> str:
        """
        Translate the Beckhoff device type code into a more suitable PV name.
        """
        if self.type == DeviceType.IODEVICETYPE_ETHERCAT:
            return f"ETH{self.id}"
        else:
            return f"EBUS{self.id}"


@dataclass
class IOServer:
    """Define an I/O server object."""

    name: str
    """The name of the server"""
    version: str
    """The version number of the server"""
    build: SupportsInt
    """The build number of the server"""
    num_devices: SupportsInt
    """The number of EtherCAT devices registered with the server"""
    category: IONodeType = IONodeType.Server
    """The component category the object belongs to in the EtherCAT system"""


class IOTreeNode:
    """
    Define an I/O component as a node in the tree structure representation \
        of the EtherCAT system.
    """

    def __init__(
        self, data: IOServer | IODevice | IOSlave, path: list[str] | None = None
    ):
        if path is None:
            self.path = [data.name]
            """The path to this node in the tree."""
        else:
            path.append(data.name)
            self.path = path

        self.data = data
        """The EtherCAT object associated to this node."""
        self.children: list[IOTreeNode] = []
        """The child nodes connected to this node."""

    @property
    def child_count(self) -> int:
        """The number of child nodes connected to this node."""
        return len(self.children)

    @property
    def tree_path(self) -> str:
        """A string representation of the path to this node in the tree."""
        assert self.path is not None
        return " <-- ".join(self.path) if len(self.path) > 1 else self.path[0]

    def add_child(self, child: Self) -> None:
        """
        Add a child node to the current node.

        :param child: the child node to add to the current node
        """
        self.children.append(child)

    def has_children(self) -> bool:
        """
        Check if the node has any child nodes.

        :returns: true if the node has one or more child nodes
        """
        return bool(self.children)

    def tree_height(self) -> int:
        """
        Compute the height of the tree starting at this node.
        The height of a tree is the number of nodes along the longest path from
        the root node down to the farthest leaf node.

        :returns: the height of the tree starting at this node"""
        if not self.children:
            return 1
        return 1 + max(child.tree_height() for child in self.children)

    def node_search(self, target: str) -> bool:
        """
        Depth-first search algorithm to check if the node is itself or has a child node\
             whose associated EtherCAT object correspond to a specific name.

        :param target: the name of the EtherCAT object to find in the tree

        :returns: true if this root node comprises the EtherCAT object name
        """
        if self.data.name == target:
            return True
        for child in self.children:
            if child.node_search(target):
                return True
        return False

    def print_tree(self, depth: int = 0) -> None:
        """
        Depth-first pre-order traversal function to print the EtherCAT object name \
            across the depth of the tree starting at this node.

        params depth: the current depth in the tree (used for indentation)
        """
        tabs = "\t" * depth
        print(f"{tabs}node: {self.data.name}")
        for child in self.children:
            child.print_tree(depth + 1)

    def node_generator(self) -> Generator["IOTreeNode", Any, Any]:
        """
        Iterate recursively over all node elements in the tree starting at this node.

        :returns: a generator yielding all nodes in the tree
        """
        yield self
        for child in self.children:
            yield from child.node_generator()



import asyncio
import random

from fastcs_catio.client import AsyncioADSClient

# Connection to CX2020 device
TARGET_IP = "172.23.240.142"
TARGET_NETID = "5.59.238.150.1.1"
TARGET_ADS_PORT = 27909


async def make_client() -> AsyncioADSClient:
    return await AsyncioADSClient.connected_to(
        target_ip=TARGET_IP,
        target_ams_net_id=TARGET_NETID,
        target_ams_port=TARGET_ADS_PORT,
    )


async def main():
    client = await make_client()

    try:
        await client.introspect_io_server()
        symbols = await client.get_all_symbols()
        master_dev = next(iter(symbols))

        # Read symbol values
        for _ in range(10):
            name, value = await client.read_ads_symbol(
                random.choice(symbols[master_dev])
            )
            print(f"{name}: {value}")

        read_symbols = random.sample(symbols[master_dev], 10)
        for symbol in read_symbols:
            name, value = await client.read_ads_symbol(symbol)
            print(f"Read {name} = {value}")

        print(f"SUM Read -- {await client.sumread_ads_symbols(read_symbols)}")

        # Write value to symbols
        output_symbols = [
            symbol
            for symbol in symbols[master_dev]
            if all(substr in symbol.name for substr in ["EL2024", "Channel"])
        ]
        if output_symbols:
            write_symbols = random.sample(
                symbols[master_dev], min(10, len(output_symbols))
            )

            for symbol in write_symbols:
                print(
                    f"{symbol.name}, "
                    + f"value set to 1: {await client.write_ads_symbol(symbol, 1)}"
                )
            await asyncio.sleep(1)
            for symbol in write_symbols:
                print(
                    f"{symbol.name}, "
                    + f"value set to 0: {await client.write_ads_symbol(symbol, 0)}"
                )

            # # SumWrite will return 'ErrorCode.ADSERR_DEVICE_SRVNOTSUPP' with CX2020.
            # # (single write command works and couldn't find any error in bytes stream)
            # n = len(write_symbols)
            # print(
            #     "SUM Write -- "
            #     + f"{await client.sumwrite_ads_symbols(
            #         list(zip(write_symbols, [1] * n)))}"
            # )
            # await asyncio.sleep(1)
            # print(
            #     "SUM Write -- "
            #     + f"{await client.sumwrite_ads_symbols(
            #         list(zip(write_symbols, [0] * n)))}"
            # )

            # # ReadWrite calls will also return 'ErrorCode.ADSERR_DEVICE_SRVNOTSUPP'.
            # # ReadWrite value to symbols
            # readwrite_symbols = random.sample(
            #     symbols[master_dev], min(5, len(output_symbols))
            # )
            # for symbol in readwrite_symbols:
            #     print(
            #         f"{symbol.name}, value set to 1: "
            #         + f"initially was {await client.readwrite_ads_symbol(symbol, 1)}"
            #     )
            # await asyncio.sleep(1)
            # for symbol in readwrite_symbols:
            #     print(
            #         f"{symbol.name}, value set to 0: "
            #         + f"initially was {await client.readwrite_ads_symbol(symbol, 0)}"
            #     )

    finally:
        await client.close()


if __name__ == "__main__":
    asyncio.run(main())
